#include <ruby.h>
#include "ruby_util.h"
#include "lexer.h"

#define YY_DECL int Lexer_yylex(TokenInfo tokenInfo, yyscan_t yyscanner)
YY_DECL;

#define YY_HEADER_EXPORT_START_CONDITIONS
#include "scanner.h"

#define Lexer_C_TOKEN_PROC(name, kind, str) { TokenKind_ ## kind, str }, 
#define Lexer_LUA_TOKEN_PROC(name, kind, str) { TokenKind_ ## kind, str },

struct {
  TokenKind kind;
  const char *str;
} pairList[] = {
  Lexer_TOKEN_LIST()
};

#undef Lexer_C_TOKEN_PROC
#undef Lexer_LUA_TOKEN_PROC

VALUE tokenList[TokenID_SIZE] = {
  Qnil
};

VALUE modLuca = Qnil;
VALUE clsLexer = Qnil;
VALUE clsToken = Qnil;
VALUE errArgumentError = Qnil;
VALUE errLexerError = Qnil;

void Lexer_yy_switch_state(int state, yyscan_t scanner);

/**
 * グローバル変数の初期化
 */
static void init_global(void)
{
  // Luca::Tokenの取得
  ID idToken = rb_intern("Token");
  clsToken = rb_const_get_at(modLuca, idToken);
  
  // ArgumentErrorの取得
  ID idArgumentError = rb_intern("ArgumentError");
  errArgumentError = rb_const_get_at(rb_cObject, idArgumentError);
  
  // Luca::LexerErrorの取得
  ID idLexerError = rb_intern("LexerError");
  errLexerError = rb_const_get_at(modLuca, idLexerError);
  
  // tokenListの初期化
  for( int i = 0; i < TokenID_SIZE; i++ ){
    VALUE token = Qnil;
    
    switch( pairList[i].kind ){
    case TokenKind_KEYWORD:
      token = rb_str_new2(pairList[i].str);
      break;
    case TokenKind_LITERAL:
      token = ID2SYM(rb_intern(pairList[i].str));
      break;
    default:
      rb_fatal("invalid pairList");
    }
    
    tokenList[i] = token;
    rb_global_variable(&(tokenList[i]));
  }
}

static VALUE Token_new(TokenKind kind, VALUE token, TokenInfo tokenInfo)
{
  VALUE line = INT2NUM(tokenInfo->line);
  VALUE value;
  VALUE newline_count = INT2NUM(tokenInfo->newline_count);
  
  if( kind == TokenKind_KEYWORD ){
    value = token;
  }
  else if( kind == TokenKind_LITERAL ){    
    value = tokenInfo->value;
  }
  
  VALUE argv[3] = { line, value, newline_count };
  
  return rb_class_new_instance(3, argv, clsToken);
}


static void Lexer_close(Lexer lexer)
{
  if( !lexer->isClosed ){
    fclose(lexer->fp);
    Lexer_yylex_destroy(lexer->scanner);
    lexer->isClosed = 1;
  }
}

static void Lexer_mark(Lexer lexer)
{
  rb_gc_mark(lexer->parser);
}

static void Lexer_free(Lexer lexer)
{
  // ファイルポインタを確実に閉じる
  Lexer_close(lexer);
  
  // initializeで複製したファイル名文字列を解放
  free(lexer->screen_filename);
}

static VALUE Lexer_alloc(VALUE cls)
{
  Lexer lexer = ALLOC(struct Lexer);
  
  return Data_Wrap_Struct(cls, Lexer_mark, Lexer_free, lexer);
}

static VALUE Lexer_initialize(VALUE self, VALUE filenamestr, VALUE screen_filenamestr, VALUE parser)
{
  Lexer lexer;
  FILE *fp;
  Data_Get_Struct(self, struct Lexer, lexer);
  
  const char *filename = StringValuePtr(filenamestr);
  if( (fp = fopen(filename, "r")) == NULL ){
    rb_raise_at(rb_eIOError, "cannot open '%s'", filename);
  }
  
  Lexer_yylex_init(&(lexer->scanner));
  lexer->fp = fp;
  lexer->isClosed = 0;
  lexer->parser = parser;
  lexer->screen_filename = strdup(StringValuePtr(screen_filenamestr));
  lexer->newline_count = 0;
  if( lexer->screen_filename == NULL ){
    rb_fatal("malloc failed");
  }
  
  yyscan_t scanner = lexer->scanner;
  
  Lexer_yyset_in(fp, scanner);
  Lexer_yyset_extra(lexer, scanner);
  Lexer_yy_push_state(LexerState_C, scanner);
  
  return self;
}

static VALUE Lexer_next_token(VALUE self)
{
  Lexer lexer;
  Data_Get_Struct(self, struct Lexer, lexer);
  
  yyscan_t scanner = lexer->scanner;
  struct TokenInfo tokenInfo;
  tokenInfo.value = Qnil;
  
  if( lexer->isClosed ){
    rb_raise(rb_eIOError, "closed stream");
  }
  
  int tokenID = Lexer_yylex(&tokenInfo, scanner);
  VALUE token;
  VALUE value;
  if( tokenID ){
    int index = tokenID - TokenID_OFFSET;
    
    token = tokenList[index];
    
    value = Token_new(pairList[index].kind, token, &tokenInfo);
  }
  else{
    token = Qfalse;
    value = rb_str_new2("$end");
    Lexer_close(lexer);
  }
  
  return rb_assoc_new(token, value);
}

static VALUE Lexer_switch_to_c(VALUE self)
{
  Lexer lexer;
  Data_Get_Struct(self, struct Lexer, lexer);
  
  yyscan_t scanner = lexer->scanner;
  Lexer_yy_push_state(LexerState_C, scanner);
  
  return Qnil;
}

static VALUE Lexer_switch_to_lua(VALUE self)
{
  Lexer lexer;
  Data_Get_Struct(self, struct Lexer, lexer);
  
  yyscan_t scanner = lexer->scanner;
  Lexer_yy_push_state(LexerState_LUA, scanner);
  
  return Qnil;
}

void Init_lexer(void)
{
  ID idLuca = rb_intern("Luca");
  modLuca = rb_const_get_at(rb_cObject, idLuca);
  
  // class Luca::Lexer
  clsLexer = rb_define_class_under(modLuca, "Lexer", rb_cObject);
  
  // def Lexer.allocate
  rb_define_alloc_func(clsLexer, Lexer_alloc);
  
  // def Lexer#initialize
  rb_define_private_method(clsLexer, "initialize", Lexer_initialize, 3);
  
  // def Lexer#next_token
  rb_define_method(clsLexer, "next_token", Lexer_next_token, 0);
  
  // def Lexer#switch_to_c
  rb_define_method(clsLexer, "switch_to_c", Lexer_switch_to_c, 0);
  
  // def Lexer#switch_to_lua
  rb_define_method(clsLexer, "switch_to_lua", Lexer_switch_to_lua, 0);
  
  // lib/lexer/token_infoをrequireして, Lexer::TokenInfoを定義
  rb_require("luca/token");
  
  // modLuca, clsLexer以外のグローバル変数の初期化
  init_global();
}
