%{
#include "lexer.h"

#include "ruby_util.h"
#include <ctype.h>

#define YY_DECL int Lexer_yylex(TokenInfo tokenInfo, yyscan_t yyscanner)

#define returnCKwd(name) \
  do{ \
    Lexer lexer = yyextra; \
    TokenInfo_set(tokenInfo, yylineno, TokenKind_KEYWORD, Qnil, lexer->newline_count); \
    lexer->newline_count = 0; \
    return TokenID_C_ ## name; \
  }while( 0 )

#define returnCLiteral(name, v) \
  do{ \
    Lexer lexer = yyextra; \
    TokenInfo_set(tokenInfo, yylineno, TokenKind_LITERAL, (v), lexer->newline_count); \
    lexer->newline_count = 0; \
    return TokenID_C_ ## name; \
  }while( 0 )

#define returnLuaKwd(name) \
  do{ \
    Lexer lexer = yyextra; \
    TokenInfo_set(tokenInfo, yylineno, TokenKind_KEYWORD, Qnil, lexer->newline_count); \
    lexer->newline_count = 0; \
    return TokenID_LUA_ ## name; \
  }while( 0 )

#define returnLuaLiteral(name, v) \
  do{ \
    Lexer lexer = yyextra; \
    TokenInfo_set(tokenInfo, yylineno, TokenKind_LITERAL, (v), lexer->newline_count); \
    lexer->newline_count = 0; \
    return TokenID_LUA_ ## name; \
  }while( 0 )


void Lexer_yy_switch_state(int state, yyscan_t scanner);

#define MSGSIZE 1024

static char *parseLuaString(const char *text, int inLen, int *outLen, char *msg);
static int isTypedefedName(Lexer lexer, const char *name);
%}

%option reentrant
%option yylineno
%option stack
%option extra-type="Lexer"

%option prefix="Lexer_yy"
%option outfile="scanner.c"
%option header="scanner.h"

%option noinput
%option nounput

%x LexerState_C
%x LexerState_LUA

NO_DIGIT [_a-zA-Z]
DIGIT [[:digit:]]
NZ_DIGIT [1-9]
OCT_DIGIT [0-7]
HEX_DIGIT [0-9a-fA-F]

DIGIT_SEQ {DIGIT}+
HEX_DIGIT_SEQ {HEX_DIGIT}+
HEX_PREFIX (0[xX])

UNSIGNED_SUFFIX [uU]
LONG_SUFFIX [lL]
LONG_LONG_SUFFIX (ll|LL)
INT_SUFFIX ({UNSIGNED_SUFFIX}{LONG_SUFFIX}?|{UNSIGNED_SUFFIX}{LONG_LONG_SUFFIX}?|{LONG_SUFFIX}{UNSIGNED_SUFFIX}?|{LONG_LONG_SUFFIX}{UNSIGNED_SUFFIX}?)

DOT [.]
SIGN [-+]
EXP_PART [eE]{SIGN}?{DIGIT_SEQ}
BIN_EXP_PART [pP]{SIGN}?{DIGIT_SEQ}
FLOAT_SUFFIX [flFL]

SIMPLE_ESC_SEQ \\['"?\\abfnrtv]
OCT_ESC_SEQ \\{OCT_DIGIT}{1,3}
HEX_ESC_SEQ \\x{HEX_DIGIT}{2,}
ESC_SEQ ({SIMPLE_ESC_SEQ}|{OCT_ESC_SEQ}|{HEX_ESC_SEQ})
C_CHAR ([^'\\\n]|{ESC_SEQ})
C_CHAR_SEQ {C_CHAR}+

S_CHAR ([^"\\\n]|{ESC_SEQ})
S_CHAR_SEQ {S_CHAR}+

%%
  // C Rule
<LexerState_C>"auto" {
  returnCKwd(AUTO);
}

<LexerState_C>"break" {
  returnCKwd(BREAK);
}

<LexerState_C>"case" {
  returnCKwd(CASE);
}

<LexerState_C>"char" {
  returnCKwd(CHAR);
}

<LexerState_C>"const" {
  returnCKwd(CONST);
}

<LexerState_C>"continue" {
  returnCKwd(CONTINUE);
}

<LexerState_C>"default" {
  returnCKwd(DEFAULT);
}

<LexerState_C>"do" {
  returnCKwd(DO);
}

<LexerState_C>"double" {
  returnCKwd(DOUBLE);
}

<LexerState_C>"else" {
  returnCKwd(ELSE);
}

<LexerState_C>"enum" {
  returnCKwd(ENUM);
}

<LexerState_C>"extern" {
  returnCKwd(EXTERN);
}

<LexerState_C>"float" {
  returnCKwd(FLOAT);
}

<LexerState_C>"for" {
  returnCKwd(FOR);
}

<LexerState_C>"goto" {
  returnCKwd(GOTO);
}

<LexerState_C>"if" {
  returnCKwd(IF);
}

<LexerState_C>"inline" {
  returnCKwd(INLINE);
}

<LexerState_C>"int" {
  returnCKwd(INT);
}

<LexerState_C>"long" {
  returnCKwd(LONG);
}

<LexerState_C>"register" {
  returnCKwd(REGISTER);
}

<LexerState_C>"restrict" {
  returnCKwd(RESTRICT);
}

<LexerState_C>"__restrict" {
  returnCKwd(US_RESCRICT);
}

<LexerState_C>"return" {
  returnCKwd(RETURN);
}

<LexerState_C>"short" {
  returnCKwd(SHORT);
}

<LexerState_C>"signed" {
  returnCKwd(SIGNED);
}

<LexerState_C>"sizeof" {
  returnCKwd(SIZEOF);
}

<LexerState_C>"static" {
  returnCKwd(STATIC);
}

<LexerState_C>"struct" {
  returnCKwd(STRUCT);
}

<LexerState_C>"switch" {
  returnCKwd(SWITCH);
}

<LexerState_C>"typedef" {
  returnCKwd(TYPEDEF);
}

<LexerState_C>"union" {
  returnCKwd(UNION);
}

<LexerState_C>"unsigned" {
  returnCKwd(UNSIGNED);
}

<LexerState_C>"void" {
  returnCKwd(VOID);
}

<LexerState_C>"volatile" {
  returnCKwd(VOLATILE);
}

<LexerState_C>"while" {
  returnCKwd(WHILE);
}

<LexerState_C>"_Bool" {
  returnCKwd(BOOL);
}

<LexerState_C>"_Complex" {
  returnCKwd(COMPLEX);
}

<LexerState_C>"_Imaginary" {
  returnCKwd(IMAGINARY);
}

<LexerState_C>"__asm__" {
  returnCKwd(ASM);
}

<LexerState_C>"__attribute__" {
  returnCKwd(ATTRIBUTE);
}

<LexerState_C>"__extension__" {
  returnCKwd(EXTENSION);
}

<LexerState_C>"[" {
  returnCKwd(L_BRACKET);
}

<LexerState_C>"]" {
  returnCKwd(R_BRACKET);
}

<LexerState_C>"(" {
  returnCKwd(L_PAR);
}

<LexerState_C>")" {
  returnCKwd(R_PAR);
}

<LexerState_C>"{" {
  returnCKwd(L_BRACE);
}

<LexerState_C>"}" {
  returnCKwd(R_BRACE);
}

<LexerState_C>"." {
  returnCKwd(DOT);
}

<LexerState_C>"->" {
  returnCKwd(ARROW);
}

<LexerState_C>"++" {
  returnCKwd(INC);
}

<LexerState_C>"--" {
  returnCKwd(DEC);
}

<LexerState_C>"&" {
  returnCKwd(AND);
}

<LexerState_C>"*" {
  returnCKwd(MUL);
}

<LexerState_C>"+" {
  returnCKwd(PLUS);
}

<LexerState_C>"-" {
  returnCKwd(MINUS);
}

<LexerState_C>"~" {
  returnCKwd(TILDE);
}

<LexerState_C>"!" {
  returnCKwd(NOT);
}

<LexerState_C>"/" {
  returnCKwd(DIV);
}

<LexerState_C>"%" {
  returnCKwd(MOD);
}

<LexerState_C>"<<" {
  returnCKwd(LSHIFT);
}

<LexerState_C>">>" {
  returnCKwd(RSHIFT);
}

<LexerState_C>"<" {
  returnCKwd(LTHAN);
}

<LexerState_C>">" {
  returnCKwd(GTHAN);
}

<LexerState_C>"<=" {
  returnCKwd(LEQ);
}

<LexerState_C>">=" {
  returnCKwd(REQ);
}

<LexerState_C>"==" {
  returnCKwd(EQL);
}

<LexerState_C>"!=" {
  returnCKwd(NEQ);
}

<LexerState_C>"^" {
  returnCKwd(XOR);
}

<LexerState_C>"||" {
  returnCKwd(LOR);
}

<LexerState_C>"&&" {
  returnCKwd(LAND);
}

<LexerState_C>"|" {
  returnCKwd(BOR);
}

<LexerState_C>"?" {
  returnCKwd(COND);
}

<LexerState_C>":" {
  returnCKwd(SELECT);
}

<LexerState_C>";" {
  returnCKwd(SEMICOLON);
}

<LexerState_C>"..." {
  returnCKwd(VARG);
}

<LexerState_C>"=" {
  returnCKwd(ASSIGN);
}

<LexerState_C>"*=" {
  returnCKwd(MUL_ASSIGN);
}

<LexerState_C>"/=" {
  returnCKwd(DIV_ASSIGN);
}

<LexerState_C>"%=" {
  returnCKwd(MOD_ASSIGN);
}

<LexerState_C>"+=" {
  returnCKwd(ADD_ASSIGN);
}

<LexerState_C>"-=" {
  returnCKwd(SUB_ASSIGN);
}

<LexerState_C>"<<=" {
  returnCKwd(LSHIFT_ASSIGN);
}

<LexerState_C>">>=" {
  returnCKwd(RSHIFT_ASSIGN);
}

<LexerState_C>"&=" {
  returnCKwd(BAND_ASSIGN);
}

<LexerState_C>"^=" {
  returnCKwd(XOR_ASSIGN);
}

<LexerState_C>"|=" {
  returnCKwd(BOR_ASSIGN);
}

<LexerState_C>"," {
  returnCKwd(COMMA);
}

<LexerState_C>"_lua" {
  returnCKwd(LUA);
}

<LexerState_C>{NO_DIGIT}({NO_DIGIT}|{DIGIT})* { // 識別子
  Lexer lexer = yyextra;
  if( isTypedefedName(lexer, yytext) ){
    returnCLiteral(TYPEDEF_NAME, rb_intern_sym(yytext));
  }
  else{
    returnCLiteral(NAME, rb_intern_sym(yytext));
  }
}

<LexerState_C>{NZ_DIGIT}{DIGIT}*{INT_SUFFIX}? { // 10進整数リテラル
  returnCLiteral(INT_LITERAL, rb_str_new2(yytext));
}

<LexerState_C>(0{OCT_DIGIT}*){INT_SUFFIX}? { // 8進整数リテラル
  returnCLiteral(INT_LITERAL, rb_str_new2(yytext));
}

<LexerState_C>({HEX_PREFIX}{HEX_DIGIT}+){INT_SUFFIX}? { //16進整数リテラル
  returnCLiteral(INT_LITERAL, rb_str_new2(yytext));
}

<LexerState_C>({DIGIT_SEQ}?{DOT}{DIGIT_SEQ}|{DIGIT_SEQ}{DOT}){EXP_PART}?{FLOAT_SUFFIX}? { // 10進浮動小数点数リテラル(小数点含み)
  returnCLiteral(FLOAT_LITERAL, rb_str_new2(yytext));
}

<LexerState_C>{DIGIT_SEQ}{EXP_PART}{FLOAT_SUFFIX}? { // 10進浮動小数点数リテラル(小数点無し)
  returnCLiteral(FLOAT_LITERAL, rb_str_new2(yytext));
}

<LexerState_C>{HEX_PREFIX}({HEX_DIGIT_SEQ}?{DOT}{HEX_DIGIT_SEQ}|{HEX_DIGIT_SEQ}{DOT}){BIN_EXP_PART}{FLOAT_SUFFIX}? {// 16進浮動小数点数リテラル(小数点含み)
  returnCLiteral(FLOAT_LITERAL, rb_str_new2(yytext));
}

<LexerState_C>{HEX_PREFIX}{HEX_DIGIT_SEQ}{BIN_EXP_PART}{FLOAT_SUFFIX}? { // 16進浮動小数点数リテラル(小数点無し)
  returnCLiteral(FLOAT_LITERAL, rb_str_new2(yytext));
}

<LexerState_C>L?[']{C_CHAR_SEQ}['] { // 文字リテラル
  returnCLiteral(CHAR_LITERAL, rb_str_new2(yytext));
}

<LexerState_C>L?["]{S_CHAR_SEQ}?["] { // 文字列リテラル
  returnCLiteral(STRING_LITERAL, rb_str_new2(yytext));
}

<LexerState_C>[ \t\r] {
  // 空白は無視
}

<LexerState_C>[\n] {
  Lexer lexer = yyextra;
  lexer->newline_count += 1;
}

<LexerState_LUA>"and" {
  returnLuaKwd(AND);
}

<LexerState_LUA>"break" {
  returnLuaKwd(BREAK);
}

<LexerState_LUA>"do" {
  returnLuaKwd(DO);
}

<LexerState_LUA>"else" {
  returnLuaKwd(ELSE);
}

<LexerState_LUA>"elseif" {
  returnLuaKwd(ELSEIF);
}

<LexerState_LUA>"end" {
  returnLuaKwd(END);
}

<LexerState_LUA>"false" {
  returnLuaKwd(FALSE);
}

<LexerState_LUA>"for" {
  returnLuaKwd(FOR);
}

<LexerState_LUA>"function" {
  returnLuaKwd(FUNCTION);
}

<LexerState_LUA>"goto" {
  returnLuaKwd(GOTO);
}

<LexerState_LUA>"if" {
  returnLuaKwd(IF);
}

<LexerState_LUA>"in" {
  returnLuaKwd(IN);
}

<LexerState_LUA>"local" {
  returnLuaKwd(LOCAL);
}

<LexerState_LUA>"nil" {
  returnLuaKwd(NIL);
}

<LexerState_LUA>"not" {
  returnLuaKwd(NOT);
}

<LexerState_LUA>"or" {
  returnLuaKwd(OR);
}

<LexerState_LUA>"repeat" {
  returnLuaKwd(REPEAT);
}

<LexerState_LUA>"return" {
  returnLuaKwd(RETURN);
}

<LexerState_LUA>"then" {
  returnLuaKwd(THEN);
}

<LexerState_LUA>"true" {
  returnLuaKwd(TRUE);
}

<LexerState_LUA>"until" {
  returnLuaKwd(UNTIL);
}

<LexerState_LUA>"while" {
  returnLuaKwd(WHILE);
}

<LexerState_LUA>"+" {
  returnLuaKwd(ADD);
}

<LexerState_LUA>"-" {
  returnLuaKwd(SUB);
}

<LexerState_LUA>"*" {
  returnLuaKwd(MUL);
}

<LexerState_LUA>"/" {
  returnLuaKwd(DIV);
}

<LexerState_LUA>"%" {
  returnLuaKwd(MOD);
}

<LexerState_LUA>"^" {
  returnLuaKwd(POW);
}

<LexerState_LUA>"#" {
  returnLuaKwd(LEN);
}

<LexerState_LUA>"==" {
  returnLuaKwd(EQL);
}

<LexerState_LUA>"~=" {
  returnLuaKwd(NEQ);
}

<LexerState_LUA>"<=" {
  returnLuaKwd(LEQ);
}

<LexerState_LUA>">=" {
  returnLuaKwd(GEQ);
}

<LexerState_LUA>"<" {
  returnLuaKwd(LTN);
}

<LexerState_LUA>">" {
  returnLuaKwd(GTN);
}

<LexerState_LUA>"=" {
  returnLuaKwd(ASSIGN);
}

<LexerState_LUA>"(" {
  returnLuaKwd(L_PAR);
}

<LexerState_LUA>")" {
  returnLuaKwd(R_PAR);
}

<LexerState_LUA>"{" {
  returnLuaKwd(L_BRACE);
}

<LexerState_LUA>"}" {
  returnLuaKwd(R_BRACE);
}

<LexerState_LUA>"[" {
  returnLuaKwd(L_BRACKET);
}

<LexerState_LUA>"]" {
  returnLuaKwd(R_BRACKET);
}

<LexerState_LUA>"::" {
  returnLuaKwd(LABEL);
}

<LexerState_LUA>":" {
  returnLuaKwd(COLON);
}

<LexerState_LUA>";" {
  returnLuaKwd(SEMICOLON);
}

<LexerState_LUA>"," {
  returnLuaKwd(COMMA);
}

<LexerState_LUA>"." {
  returnLuaKwd(DOT);
}

<LexerState_LUA>".." {
  returnLuaKwd(APPEND);
}

<LexerState_LUA>"..." {
  returnLuaKwd(VARG);
}

<LexerState_LUA>"_clocal" {
  returnLuaKwd(C_LOCAL);
}

<LexerState_LUA>"_protected" {
  returnLuaKwd(PROTECTED);
}

<LexerState_LUA>"_setmetatable" {
  returnLuaKwd(SETMETATABLE);
}

<LexerState_LUA>"`" {
  returnLuaKwd(BACK_QUOTE);
}

<LexerState_LUA>{NO_DIGIT}({DIGIT}|{NO_DIGIT})* { // 識別子
  returnLuaLiteral(NAME, rb_intern_sym(yytext));
}

<LexerState_LUA>({NZ_DIGIT}{DIGIT}*|0) { // 10進整数リテラル
  long n;
  sscanf(yytext, "%ld", &n);
  returnLuaLiteral(INT_LITERAL, LONG2NUM(n));
}

<LexerState_LUA>({HEX_PREFIX}{HEX_DIGIT}+) { //16進整数リテラル
  long n;
  sscanf(yytext, "%lx", &n);
  returnLuaLiteral(INT_LITERAL, LONG2NUM(n));
}

<LexerState_LUA>({DIGIT_SEQ}?{DOT}{DIGIT_SEQ}|{DIGIT_SEQ}{DOT}){EXP_PART}? { // 10進浮動小数点数リテラル(小数点含み)
  double n;
  sscanf(yytext, "%lf", &n);
  returnLuaLiteral(FLOAT_LITERAL, DBL2NUM(n));
}

<LexerState_LUA>{DIGIT_SEQ}{EXP_PART} { // 10進浮動小数点数リテラル(小数点無し)
  double n;
  sscanf(yytext, "%lf", &n);
  returnLuaLiteral(FLOAT_LITERAL, DBL2NUM(n));
}

<LexerState_LUA>{HEX_PREFIX}({HEX_DIGIT_SEQ}?{DOT}{HEX_DIGIT_SEQ}|{HEX_DIGIT_SEQ}{DOT}){BIN_EXP_PART} {// 16進浮動小数点数リテラル(小数点含み)
  double n;
  sscanf(yytext, "%la", &n);
  returnLuaLiteral(FLOAT_LITERAL, DBL2NUM(n));
}

<LexerState_LUA>{HEX_PREFIX}{HEX_DIGIT_SEQ}{BIN_EXP_PART} { // 16進浮動小数点数リテラル(小数点無し)
  double n;
  sscanf(yytext, "%la", &n);
  returnLuaLiteral(FLOAT_LITERAL, DBL2NUM(n));
}

<LexerState_LUA>['](\\(.)|[^'\n]*)['] { // 文字列リテラル (シングルクォート)
  int len;
  char msg[MSGSIZE];

  char *result = parseLuaString(yytext, yyleng, &len, msg);


  if( result == NULL ){
    rb_raise(rb_eStandardError, "line %d, invalid string literal %s: %s", yylineno, yytext, msg);
  }

  VALUE str = rb_str_new(result, len);
  free(result);

  returnLuaLiteral(STRING_LITERAL, str);
}

<LexerState_LUA>["](\\(.)|[^"\n]*)["] { // 文字列リテラル (ダブルクォート)
  int len;
  char msg[MSGSIZE];

  char *result = parseLuaString(yytext, yyleng, &len, msg);


  if( result == NULL ){
    Lexer lexer = yyextra;
    rb_raise(errLexerError, "%s:%d: invalid string literal %s: %s", lexer->screen_filename, yylineno, yytext, msg);
  }
  VALUE str = rb_str_new(result, len);
  free(result);

  returnLuaLiteral(STRING_LITERAL, str);
}

<LexerState_LUA>[ \t\r] {} // 空白

<LexerState_LUA>[\n] { // 改行
  Lexer lexer = yyextra;
  lexer->newline_count += 1;
}

<LexerState_C,LexerState_LUA>. { // 解析不可能な文字
  Lexer lexer = yyextra;

  rb_raise(errLexerError, "%s:%d: unreconized character `%s'", lexer->screen_filename, yylineno, yytext);
}

<INITIAL>. {
  rb_raise_at(rb_eStandardError, "lexer is INITIAL state (this is BUG!)");
}

%%

int Lexer_yywrap(yyscan_t scanner)
{
  return 1;
}

void Lexer_yy_push_state(int state, yyscan_t scanner)
{
  yy_push_state(state, scanner);
}

void Lexer_yy_pop_state(yyscan_t scanner)
{
  yy_pop_state(scanner);
}

void Lexer_yy_switch_state(int state, yyscan_t scanner)
{
  printf("begin switch to %d\n", state);

  yy_pop_state(scanner);
  yy_push_state(state, scanner);

  printf("end switch to %d\n", state);
}

int Lexer_yy_top_state(yyscan_t scanner)
{
  return yy_top_state(scanner);
}

static char *parseLuaString(const char *text, int inLen, int *outLen, char *msg)
{
  char *ret = malloc(inLen + 1);
  inLen -= 2; // 前後のダブルクォートの長さを除去
  if( ret == NULL ){
    snprintf(msg, MSGSIZE, "malloc failed");
    return NULL;
  }

  // 結果の文字書き込み先ポインタ
  char *s = ret;
  // 結果の文字列の長さ
  int len = 0;
  // 1から始めるのは始めのダブルクォートを読み飛ばす為
  for( int i = 1; i < inLen + 1; i++, s++ ){
    char ch;
    if( text[i] == '\\' ){
      // エスケープシーケンスの場合

      i++; // 次の文字へ進む
      if( isdigit(text[i]) ){
        // 10進数定数バイト
        int tmp;
        int ch_len = 1;
        char buf[4] = { text[i], '\0', '\0', '\0' };

        for( int j = i+1; j - i < 3; j++ ){
          if( !isdigit(text[j]) ) break;
          buf[j-i] = text[j];
          ch_len++;
        }

        tmp =strtol(buf, NULL, 10);
        if( tmp > 0xff ){
          snprintf(msg, MSGSIZE, "too many decimal character %d", tmp);
          return NULL;
        }

        i += ch_len;
        ch = tmp;
      }
      else{
        switch( text[i] ){
        case 'a':
          ch = '\a';
          break;

        case 'b':
          ch = '\b';
          break;

        case 'f':
          ch = '\f';
          break;

        case 'n':
          ch = '\n';
          break;

        case 'r':
          ch = '\r';
          break;

        case 't':
          ch = '\t';
          break;

        case 'v':
          ch = '\v';
          break;

        case '\\':
          ch = '\\';
          break;

        case '\'':
          ch = '\'';
          break;

        case '\"':
          ch = '\"';
          break;

        case 'x': {
          // 16進数定数バイト
          char buf[3];

          if( !isxdigit(text[i+1]) || !isxdigit(text[i+2]) ){
            snprintf(msg, MSGSIZE, "illegal hexmal character");
            return NULL;
          }

          buf[0] = text[i+1];
          buf[1] = text[i+2];
          buf[2] = '\0';
          i += 2;
          ch = strtol(buf, NULL, 16);
          break;
        }

        default:
          snprintf(msg, MSGSIZE, "invalid escape sequence '\\%c'\n", text[i]);
          return NULL;
        }
      }
    }
    else if( isprint(text[i]) ){
      ch = text[i];
    }
    else{
      snprintf(msg, MSGSIZE, "invalid byte 0x%x", (int)text[i]);
      return NULL;
    }

    *s = ch;
    len++;
  }

  if( outLen != NULL ){
    *outLen = len;
  }
  ret[len] = '\0';
  return ret;
}

static int isTypedefedName(Lexer lexer, const char *name)
{
  VALUE parser = lexer->parser;

  return RTEST(rb_call_method(parser, "is_typedefed_name?", 1, rb_intern_sym(name)));
}
