# rubocop:disable Style/FileName

# ffi-clangのtypeを拡張
# 各種述語などを定義

require 'ffi/clang'

module FFI
  module Clang
    class Type
      include Enumerable
      
      # @return [[Symbol]] {#kind}が返す種類のうち, 非公開な型のリスト
      INT_TYPE_LIST = %i(
        type_bool
        type_char_u
        type_uchar
        type_char16
        type_char32
        type_ushort
        type_uint
        type_ulong
        type_ulonglong
        type_uint128
        type_char_s
        type_schar
        type_wchar
        type_short
        type_int
        type_long
        type_longlong
        type_int128
        type_enum
      )
        
      # selfが整数型か否かを返す
      #
      #
      # @return [Boolean]
      # 
      def integer?
        t = canonical_rec
        
        INT_TYPE_LIST.member?(t.kind)
      end
      
      def int?
        canonical_rec.kind == :type_int
      end
      
      # selfが<tt>long</tt>型か否かを返す
      #
      #
      # @return [Boolean]
      # 
      def long?
        canonical_rec.kind == :type_long
      end
      
      FLOAT_POINT_TYPE_LIST = %i(type_float type_double type_longdouble)

      
      # selfが浮動小数点数型か否かを返す
      #
      #
      # @return [Boolean]
      # 
      def float_point?
        t = canonical_rec
        
        FLOAT_POINT_TYPE_LIST.member?(t.kind)
      end
      
      def number?
        integer? || float_point?
      end
      
      # selfがポインタか否かを返す
      #
      #
      # @return [Boolean]
      # 
      def pointer?
        canonical_rec.kind == :type_pointer
      end
      
      # selfがC文字列(<tt>char *</tt>)か否かを返す
      #
      #
      # @return [Boolean]
      # 
      def string?
        t = canonical_rec
        t.pointer? && t.pointee.kind == :type_char_s
      end
      
      # selfが関数型か否かを返す
      #
      #
      # @return [Boolean]
      # 
      def function?
        t = canonical_rec
        %i(type_function_proto type_function_noproto).member?(t.kind)
      end
      
      # selfが関数ポインタか否かを返す
      #
      #
      # @return [Boolean]
      # 
      def function_pointer?
        t = canonical_rec
        t.pointer? && t.pointee.function?
      end
      
      # selfが<tt>void</tt>か否かを返す
      #
      #
      # @return [Boolean]
      # 
      def void?
        t = self.canonical_rec
        t.kind == :type_void
      end
      
      # selfが<tt>void *</tt>か否かを返す. 
      #
      #
      # @return [Boolean]
      # 
      def void_pointer?
        t = self.canonical_rec
        t.pointer? && t.pointee.void?
      end
      
      # @return [[Symbol]] {#kind}が返す種類のうち, 非公開な型のリスト
      UNEXPOSED_TYPE_LIST = [:type_typedef, :type_unexposed]
      
      # unexposedやtypedefされている場合, 実体が現れるまで宣言をたどり続ける
      #
      #
      # @return [Type]
      # 
      def canonical_rec
        ret = self
        
        loop do
          return ret unless UNEXPOSED_TYPE_LIST.member?(ret.kind)
          ret = ret.canonical
        end
      end
      
      # 引数の型を列挙する
      #
      #
      # @return [self]
      # 
      def each_arg_type
        fail 'attempt each_arg_type for not function type' unless function?
        
        if block_given?
          num_arg_types.times do |i|
            yield arg_type(i)
          end
          self
        else
          Enumerator.new(num_arg_types) do |y|
            num_arg_types.times do |i|
              y << arg_type(i)
            end
            self
          end
        end
      end
      alias_method :each, :each_arg_type
      
      # 名前が<tt>name</tt>, 型がselfとなるような宣言文字列を生成
      #
      # @param [String, Symbol] name 宣言する変数名
      #
      # @return [String]
      # 
      def gen_decl(name)
        t = self.canonical_rec
        name = name.to_s
        
        if t.function_pointer?
          ft = t.pointee
          fail NotImplementedError, 'function of function type is not impelemeted' if ft.result_type.function?
          
          ret = ft.result_type.spelling + " (*#{name})"
          arg = ft.map(&:spelling).join(', ')
          arg << ', ...' if ft.num_arg_types > 1 && ft.variadic?
          
          ret << '(' << arg << ')'
        elsif t.pointer?
          t.pointee.spelling + " (*#{name})"
        else
          t.spelling + ' ' + name
        end
      end
    end
  end
end
