
module Luca
  include Varspec::BuiltinMatcher
  
  # AST上の意味解析を行うためのクラス
  # 
  # 解析は複数回のパスによって実現される. 
  # 個々の解析はAnalyzer以下にあるクラスによって行われる. 
  # 
  #   - 全てのノードに親ノードを付与する
  #   - チャンク(<tt>_lua</tt>文とその中のLua関数)を集める {ChunkCollecter}
  #   
  #   - バッククォート 
  #
  # @author autopp <autopp.inc@gmail.com>
  #
  class Analyzer
    # @return [Compiler] この解析器を利用するコンパイラ
    attr_reader :compiler
    
    def initialize(compiler)
      expect_var[:compiler].of_instance_to_be Compiler
    end
    
    # プログラム全体を解析する. 
    # 解析にはASTの変換を伴う可能性がある. 
    # 
    # 
    #
    # @param [Ast::C::Program] program 解析対象のAST
    #
    # @return [Ast::C::Program] 解析後のAST
    # 
    def analyze(program)
      expect_var[:program].to_be Ast::C::Program
      
      # Cコード部分のチェック
      c_code_checker = CCodeChecker.new(compiler)
      c_code_checker.check_program(program, compiler.filename)
      
      # 全てのノードに親ノードを付与する
      parent_setter = ParentSetter.new
      parent_setter.set_parent_for(program)
      
      # Luaプログラム, 関数, ラベルの収集
      chunk_collector = ChunkCollector.new
      chunk_collector.collect_chunk(program)
      
      return program if program.lua_stmt_list.empty?
      
      # Luaのジャンプに関する解析
      jump_analyzer = JumpAnalyzer.new
      jump_analyzer.analyze(program)
      
      # Luaのスコープに関する解析
      scope_analyzer = ScopeAnalyzer.new
      program = scope_analyzer.analyze(program)
      
      # 型検査
      type_checker = TypeChecker.new(compiler)
      type_checker.check_type(program, compiler.filename)
      
      # スタック解析
      stack_analyzer = StackAnalyzer.new
      stack_analyzer.analyze(program)
      
      program
    end
  end
end

require 'luca/analyzer/c_code_checker'
require 'luca/analyzer/parent_setter'
require 'luca/analyzer/chunk_collector'
require 'luca/analyzer/jump_analyzer'
require 'luca/analyzer/scope_analyzer'
require 'luca/analyzer/type_checker'
require 'luca/analyzer/stack_analyzer'
