require 'set'

module Luca
  class Analyzer
    # スコープ解析器
    #
    # @author autopp <autopp.inc@gmail.com>
    #
    class ScopeAnalyzer
      include Varspec::BuiltinMatcher

      VarInfo = Struct.new('VarInfo', :name, :id)

      # @return [Integer]
      attr_reader :function_id

      attr_reader :escape_var_count

      # programをスコープ解析する
      #
      # @param [Ast::C::Program] program 解析対象のAST
      #
      # @return [program]
      #
      def analyze(program)
        expect_var[:program].to_be Ast::C::Program

        @function_stack = []
        @function_id = 0
        @local_var_count = 0
        @escape_var_set = Set.new

        program.lua_stmt_list.map! do |lua_stmt|
          analyze_ast(lua_stmt.lua_program)
        end

        program.escape_var_set = @escape_var_set

        program
      end

      def analyze_ast(ast)
        expect_var[:ast].to_be Or(Ast, Token)
        ret = ast.analyze_scope_by(self)
        expect_var[:ret].to_be Or(Ast, Token)

        ret
      end

      def analyze_list(list)
        expect_var[:list].to_be ArrayOf(Or(Ast, Token))

        list.map do |ast|
          analyze_ast(ast)
        end
      end

      def current_function
        @function_stack.last
      end

      def enclosure_function
        @function_stack[-2].function
      end

      def current_function_id_list
        @function_stack.map(&:id)
      end

      def nest_function(function, var_list, varg)
        var_list = var_list.map.with_index { |v, i| VarInfo.new(v, @local_var_count + i) }
        @local_var_count += var_list.size
        prev_escape_var_count = @escape_var_count
        @escape_var_count = 0
        @function_stack.push(FunctionFrame.new(function, var_list, varg, @function_id))
        @function_id += 1
        ret = yield
        @escape_var_count = prev_escape_var_count
        @function_stack.pop
        ret
      end

      # Luaのローカル変数を追加
      #
      # @param [[Symbol]] var_list
      #
      # @return [object] ブロックの実行結果
      #
      def nest_lua_local_scope(var_list, &block)
        var_list = var_list.map.with_index { |v, i| VarInfo.new(v, @local_var_count + i) }
        @local_var_count += var_list.size
        current_function.nest_lua_local_scope(var_list, &block)
      end

      # Luaの<tt>protected</tt>変数スコープを追加した環境でブロックを実行する
      #
      # @param [[Symbol]] var_list 変数名のリスト
      #
      # @return [object] ブロックの実行結果
      #
      # @yield スコープを追加した環境で実行したいコード
      #
      def nest_protected_scope(var_list, &block)
        var_list = var_list.map { |v| VarInfo.new(v, nil) }
        current_function.nest_protected_scope(var_list, &block)
      end

      def nest_c_local_scope(var_list, &block)
        current_function.nest_c_local_scope(var_list, &block)
      end

      # 現在のコンテキストにおいてLua変数名がどのレベルに存在するかを返す
      #
      # @param [Symbol] name lookupする名前
      #
      # @return [[:local, [Integer, Integer], object]] ローカル変数の場合．
      #   第2要素はいくつ外側の関数か(0ならこの関数), いくつめのローカル変数かをペアにした配列．
      #   第3要素はID
      # @return [[:protected, nil]] <tt>protected</tt>変数の場合
      # @return [[:global, nil]] グローバル変数の場合
      #
      def lookup_lua_var(name)
        @function_stack.reverse_each.with_index do |frame, i|
          frame.scope_stack.reverse_each do |scope|
            next if scope.is_a?(CLocalScopeFrame)

            idx = scope.var_list.find_index { |var_info| var_info.name == name }

            next unless idx

            case scope
            when LuaLocalScopeFrame
              id = scope.var_list[idx].id
              # 外側を参照していたらエスケープセットに追加
              if i > 0
                @escape_var_set << id
                @escape_var_count += 1
              end
              return [Ast::Lua::Var::LOCAL, [i, scope.offset + idx], id]
            when ProtectedScopeFrame
              return [Ast::Lua::Var::PROTECTED, nil]
            end
          end
        end

        # 見つからなかった場合
        [Ast::Lua::Var::GLOBAL, nil]
      end

      # 現在のコンテキストにおいて，C変数が<tt>clocal</tt>宣言されている位置を返す
      #
      # @param [Symbol] name C変数名
      #
      # @return [[Integer, Integer, Integer]] アクセスパス + 宣言元のID
      # @return [nil] 通常のC変数の場合
      #
      def lookup_c_var(name)
        expect_var[:name].to_be Symbol
        @function_stack.reverse_each.with_index do |frame, i|
          frame.scope_stack.reverse_each do |scope|
            next unless scope.is_a?(CLocalScopeFrame)

            idx = scope.var_list.find_index(name)
            next unless idx

            return [@function_stack.size - i - 1, scope.offset + idx, frame.id]
          end
        end

        nil
      end

      # 関数フレーム
      #
      # @author autopp <autopp.inc@gmail.com>
      #
      class FunctionFrame
        include Varspec::BuiltinMatcher

        attr_reader :function

        attr_reader :scope_stack

        attr_reader :current_lua_offset

        # @return [Integer]
        attr_reader :id

        def initialize(function, param_list, varg, id)
          expect_var[:function].of_instance_to_be Or(Ast::Lua::FunctionLiteral, Ast::Lua::Program)
          expect_var[:param_list].to_be ArrayOf(VarInfo)
          expect_var[:varg].to_be Boolean
          expect_var[:id].of_instance_to_be Integer
          @scope_stack = [LuaLocalScopeFrame.new(param_list, 0)]
          @current_lua_offset = param_list.size
          @current_c_offset = 0
          @varg = varg
        end

        def varg?
          @varg == true
        end

        def nest_lua_local_scope(var_list)
          @scope_stack.push(LuaLocalScopeFrame.new(var_list, @current_lua_offset))
          @current_lua_offset += var_list.size
          ret = yield
          @scope_stack.pop
          ret
        end

        #
        #
        # @param [[Symbol]] var_list
        #
        # @return [object] ブロックの実行結果
        #
        def nest_protected_scope(var_list)
          @scope_stack.push(ProtectedScopeFrame.new(var_list))
          ret = yield
          @scope_stack.pop
          ret
        end

        def nest_c_local_scope(var_list)
          @scope_stack.push(CLocalScopeFrame.new(var_list, @current_c_offset))
          @current_c_offset += var_list.size
          ret = yield
          @scope_stack.pop
          ret
        end
      end

      class ScopeFrame
        include Varspec::BuiltinMatcher

        attr_reader :var_list

        def initialize(var_list)
          fail "#{self.class} is abstract class" if self.class == ScopeFrame
          expect_var[:var_list].of_instance_to_be Array
        end
      end

      # ローカル変数のフレームを表すクラス
      # @note このクラスは抽象クラスである
      #
      # @author autopp <autopp.inc@gmail.com>
      #
      class LocalScopeFrame < ScopeFrame
        # @return [Integer]
        attr_reader :offset

        #
        #
        # @param [[Symbol]] var_list
        # @param [Integer] offset
        #
        def initialize(var_list, offset)
          fail "#{self.class} is abstract class" if self.class == LocalScopeFrame

          expect_var[:offset].of_instance_to_be Integer
          expect_var[:var_list].to_be Array
          super(var_list)
        end
      end

      class LuaLocalScopeFrame < LocalScopeFrame; end

      # <tt>protected</tt>変数のフレームを表すクラス
      #
      # @author autopp <autopp.inc@gmail.com>
      #
      class ProtectedScopeFrame < ScopeFrame
        #
        #
        # @param [[Symbol]] var_list
        #
        def initialize(var_list)
          super(var_list)
        end
      end

      class CLocalScopeFrame < LocalScopeFrame; end
    end
  end
end
