require 'ffi/clang'

module Luca
  class Analyzer
    # <tt>_lua</tt>文内のバッククォートされたC式の型を検査するクラス
    #
    # ffi-clangを利用する
    #
    # @author autopp <autopp.inc@gmail.com>
    #
    class TypeChecker
      include Varspec::BuiltinMatcher
      include DebugUtils

      # @return [Compiler] この型検査器を生成したコンパイラ
      attr_reader :compiler

      def initialize(compiler)
        expect_var[:compiler].of_instance_to_be Compiler
      end

      def check_type(program, filename = '')
        expect_var[:program].to_be Ast::C::Program
        expect_var[:filename].to_be String

        # clang用の入力を生成
        input_name = "#{filename}.clang_input.c"
        File.open(input_name, 'w') do |f|
          input_generator = ClangInputGenerator.new(filename, f, 2, compiler.loc_map)
          input_generator.gen_input(program)
        end

        # 翻訳単位を作成
        tu = FFI::Clang::Index.new.parse_translation_unit(input_name, ['-std=c99'])

        # <tt>_lua</tt>文のstate式の型をチェック
        visit_lua_stmt(tu, program)

        # バッククォート式を解析
        visit_bquote(tu, program)

        # 解析漏れのあるバッククォート式がないかチェック
        program.bquote_list.each do |bquote|
          unless bquote.type
            compiler.add_error(bquote.start_pos, 'missing typing for back quote expression')
          end
        end

        # <tt>clocal</tt>宣言の型をチェック
        visit_c_local(tu, program)
        program.clocal_decl_list.each do |c_local|
          unless c_local.type
            compiler.add_error(c_local.start_pos, 'missing typing for clocal declaration')
          end
        end

        FileUtils.rm(input_name)

        program
      end

      LUA_STMT_LABEL_PREFIX = 'LuCa_LUA_STMT'
      def self.lua_stmt_label_prefix
        LUA_STMT_LABEL_PREFIX
      end

      BQUOTE_LABEL_PREFIX = 'LuCa_BQUOTE'
      def self.bquote_label_prefix
        BQUOTE_LABEL_PREFIX
      end

      C_LOCAL_LABEL_PREFIX = 'LuCa_CLOCAL'
      def self.c_local_label_prefix
        C_LOCAL_LABEL_PREFIX
      end

      private

      # 以下, clangを用いたAST解析用ヘルパ

      # ffi-clangで作成した翻訳単位を用いて, <tt>_lua</tt>文のstate式を解析する
      #
      # @param [FFI::Clang::TranslationUnit] tu 利用する翻訳単位
      # @param [Ast::C::Program] program LuCa上のAST
      #
      # @return [Ast::C::Program] programを返す
      #
      def visit_lua_stmt(tu, program)
        # 状態遷移しながらClangASTを走査
        after_label = false
        current_lua_stmt = nil
        count = 0
        cursor = tu.cursor
        cursor.visit_children do |cur, _|
          if after_label
            count += 1
            type = cur.type
            type = type.canonical if type.kind == :type_unexposed

            if type.spelling != 'lua_State *'
              compiler.add_error(current_lua_stmt.state_expr.start_pos, "expect state expr to be 'lua_State *', but got #{type.spelling}")
            end
            current_lua_stmt = nil
            after_label = false
          else
            if cur.kind == :cursor_label_stmt && cur.display_name =~ /#{LUA_STMT_LABEL_PREFIX}(\d+)/
              current_lua_stmt = program.lua_stmt_list[Regexp.last_match[1].to_i]
              after_label = true
            end
          end
          :recurse
        end
        # 解析漏れが無いかチェック
        size = program.lua_stmt_list.size
        summary = "#{count}/#{size}"

        compiler.add_error(nil, "missing type checking for state expr (#{summary})") if count != size
      end

      # ffi-clangで作成した翻訳単位を用いて, バッククォート式を解析する
      #
      # @param [FFI::Clang::TranslationUnit] tu 利用する翻訳単位
      # @param [Ast::C::Program] program LuCa上のAST
      #
      # @return [Ast::C::Program] programを返す
      #
      def visit_bquote(tu, program)
        # 状態遷移しながらClangASTを走査
        after_label = false
        current_bquote = nil
        cursor = tu.cursor
        cursor.visit_children do |cur, _|
          if after_label
            type = cur.type
            type = type.canonical if type.kind == :type_unexposed
            current_bquote.type = type
            current_bquote = nil
            after_label = false
          else
            if cur.kind == :cursor_label_stmt && cur.display_name =~ /#{BQUOTE_LABEL_PREFIX}(\d+)/
              current_bquote = program.bquote_list[Regexp.last_match[1].to_i]
              after_label = true
            end
          end
          :recurse
        end
      end

      def visit_c_local(tu, program)
        # 状態遷移しながらClangASTを走査
        after_label = false
        current_clocal = nil
        cursor = tu.cursor
        cursor.visit_children do |cur, _|
          if after_label
            type = cur.type
            type = type.canonical if type.kind == :type_unexposed
            current_clocal.type = type
            current_clocal = nil
            after_label = false
          else
            if cur.kind == :cursor_label_stmt && cur.display_name =~ /#{C_LOCAL_LABEL_PREFIX}(\d+)/
              current_clocal = program.clocal_decl_list[Regexp.last_match[1].to_i]
              after_label = true
            end
          end
          :recurse
        end
      end
    end
  end
end

require 'luca/analyzer/type_checker/clang_input_generator'
