module Luca
  # 全てのLuCaのASTノードの親クラス
  #
  # @author autopp <autopp.inc@gmail.com>
  #
  class Ast
    include Varspec::BuiltinMatcher
    
    attr_reader :start_pos
    attr_reader :end_pos
    
    def initialize(start_pos, end_pos)
      fail "#{self.class} is abstract class" if self.class == Ast
      
      @start_pos, @end_pos = expect_var[:start_pos, :end_pos].to_be Integer
    end
    
    # このクラスが抽象クラスか否かを返す
    #
    #
    # @return [Boolean] 
    # 
    def self.abstract?
      !!@is_abstract
    end
    
    # このクラスが表すノードが糖衣構文か否かを返す
    #
    #
    # @return [Boolean] 
    # 
    def self.syntax_sugar?
      !!@is_syntax_sugar
    end
    
    def inspect
      "#<#{self.class}: #{start_pos}-#{end_pos}>"
    end
    
    # <tt>cxt</tt>を用いてASTを印字する
    # 
    # @note これは抽象メソッドである. 派生クラスはこのメソッドを定義する必要がある. 
    #
    # @param [Printer] cxt 
    # 
    def print_by(cxt)
      fail "#{self.class}#print_by is not implemented"
    end
    
    def enter_desugaring_by(cxt)
      self.desugar_by(cxt)
    end
    
    # <tt>cxt</tt>を用いてASTをdesugarする
    # 
    # @note これは抽象メソッドである. 派生クラスはこのメソッドを定義する必要がある. 
    #
    # @param [DesugarContext] cxt 
    #
    # @return [Ast] desugar済みのAST
    # 
    def desugar_by(cxt)
      fail "#{self.class}#desugar_by is not implemented"
    end
    
    # <tt>name_list</tt>で指定したインスタンス変数をdesguarし, 書き換える
    # 
    # インスタンス変数が{Ast}なら, {DesugarContext#desugar_ast}を使ってdesugarする
    # 
    # インスタンス変数が{Ast}の配列なら, {DesugarContext#desugar_list}を使ってdesugarする
    # 
    # @param [DesugarContext] cxt 
    # @param [[Symbol, String]] name_list インスタンス変数名
    # 
    # @return [object, Array] desugarしたインスタンス変数
    # 
    def desugar_ivar_by(cxt, *name_list)
      name_list.each do |name|
        v = self.instance_variable_get(name)
        case
        when v.is_a?(Ast)
          v = cxt.desugar_ast(v)
        when v.is_a?(Array)
          v = cxt.desugar_list(v)
        else
          fail TypeError, "require Ast or [Ast], but got #{v.inspect}"
        end
        
        self.instance_variable_set(name, v)
      end
      
      self
    end
    
    def set_parent_by(cxt)
      fail NotImplementedError, "#{self.class}#set_parent_by is not implemented"
    end
    
    def parent
      fail 'no setting for @parent' unless instance_variable_defined?(:@parent)
      @parent
    end
    
    # 親ノードを設定する
    # 
    # 設定は1回しか行えない. 
    # 
    # @param [Ast, nil] parent 設定する親ノード
    #
    # @return [Ast, nil] parentを返す
    # 
    def parent=(parent)
      fail 'multiple parent setting' if self.instance_variable_defined?(:@parent)
      @parent = parent
    end
    
    def analyze_scope_of_ivar_by(cxt, *name_list)
      expect_var[:cxt].to_be Analyzer::ScopeAnalyzer
      name_list.each do |name|
        v = self.instance_variable_get(name)
        case
        when v.is_a?(Ast)
          v = cxt.analyze_ast(v)
        when v.is_a?(Array)
          v = cxt.analyze_list(v)
        else
          fail TypeError, "require Ast or [Ast], but got #{v.inspect}"
        end
        
        self.instance_variable_set(name, v)
      end
      
      self
    end
    
    def gen_input_by(cxt)
      fail_by_abstract_method
    end
    
    def gen_code_by(cxt)
      fail_by_abstract_method
    end
    
    private
    
    def fail_by_abstract_method
      name = caller_locations.first.base_label
      fail NotImplementedError, "#{self.class}##{name} is not implemented (line: #{start_pos}-#{end_pos})"
    end
  end
end

require 'luca/ast/c'
require 'luca/ast/lua'
