module Luca
  class Ast
    # CのASTノード
    # 
    # @note このクラスは抽象クラスである
    #
    # @author autopp <autopp.inc@gmail.com>
    #
    class C < Ast
      def desugar_by(cxt)
        self
      end
      
      # 
      # ASTのLua部分について, 再帰的に変換を行う. 
      # 変換は<tt>method</tt>によって行われ, <tt>args</tt>と(与えられば)<tt>block</tt>はそのメソッドへと渡される. 
      # 
      # CのASTノードはLuaプログラムが現れる可能性がある部分のみを再帰する. 
      # 
      # 各クラスはLua部分を再帰するために{#perform_translate_lua_by}を実装する必要がある. 
      # overrideに真が指定され, かつ<tt>method</tt>が定義されていれば, 代わりに<tt>method</tt>が呼び出される. 
      # この時, 引数として<tt>recv</tt>と<tt>args</tt>, そしてブロックとして<tt>block</tt>が渡される. 
      # 
      # @param [Symbol] recv 変換用のメソッド名
      # @param [Symbol] method 変換用の
      # 
      def translate_lua_by(recv, method, *args, override: false, &block)
        if override && self.class.instance_methods(false).member?(method)
          self.send(method, recv, *args, &block)
        else
          perform_translate_lua_by(recv, method, *args, override: override, &block)
        end
      end
      
      def perform_translate_lua_by(recv, method, *args, override: false, &block)
        self
      end
      protected :perform_translate_lua_by
      
      # 
      # ASTのLua部分について, 再帰的に走査する. 
      # 走査は<tt>method</tt>によって行われ, <tt>args</tt>と(与えられば)<tt>block</tt>はそのメソッドへと渡される. 
      # 
      # CのASTノードはLuaプログラムが現れる可能性がある部分のみを再帰する. 
      # 
      def traverse_lua_by(recv, method, *args, override: false, &block)
        if override && self.class.instance_methods(false).member?(method)
          self.send(method, recv, *args, &block)
        else
          perform_traverse_lua_by(recv, method, *args, override: override, &block)
        end
      end
      
      def perform_traverse_lua_by(recv, method, *args, override: false, &block)
        nil
      end
      protected :perform_traverse_lua_by
      
      def print_c_by(cxt)
        print_by(cxt)
      end
      
      def gen_input_by(cxt)
        print_by(cxt)
      end
      
      def gen_code_by(cxt)
        print_by(cxt)
      end
    end
  end
end

require 'luca/ast/c/expr'
require 'luca/ast/c/literal'
require 'luca/ast/c/op_node'
require 'luca/ast/c/bin_op_node'
require 'luca/ast/c/un_op_node'
require 'luca/ast/c/prefix_op_node'
require 'luca/ast/c/postfix_op_node'
require 'luca/ast/c/stmt'
require 'luca/ast/c/labeled_stmt'
require 'luca/ast/c/case_stmt'
require 'luca/ast/c/default_stmt'
require 'luca/ast/c/expr_stmt'
require 'luca/ast/c/empty_stmt'
require 'luca/ast/c/if_stmt'
require 'luca/ast/c/switch_stmt'
require 'luca/ast/c/while_stmt'
require 'luca/ast/c/do_while_stmt'
require 'luca/ast/c/for_stmt'
require 'luca/ast/c/goto_stmt'
require 'luca/ast/c/continue_stmt'
require 'luca/ast/c/break_stmt'
require 'luca/ast/c/return_stmt'
require 'luca/ast/c/compound_stmt'
require 'luca/ast/c/conditional_op_node'
require 'luca/ast/c/cast_expr'
require 'luca/ast/c/size_of_type_expr'
require 'luca/ast/c/subscript_expr'
require 'luca/ast/c/member_access_expr'
# require 'luca/ast/c/dot_access_expr'
# require 'luca/ast/c/arrow_access_expr'
require 'luca/ast/c/call_expr'
require 'luca/ast/c/var'
require 'luca/ast/c/parentheses_expr'
require 'luca/ast/c/decl_specifier'
require 'luca/ast/c/type_specifier'
require 'luca/ast/c/builtin_type_specifier'
require 'luca/ast/c/typedefed_type_specifier'
require 'luca/ast/c/user_type_specifier'
require 'luca/ast/c/enum_type_specifier'
require 'luca/ast/c/enum_constant'
require 'luca/ast/c/membered_type_specifier'
require 'luca/ast/c/member_decl'
require 'luca/ast/c/type_qualifier'
require 'luca/ast/c/const_qualifier'
require 'luca/ast/c/restrict_qualifier'
require 'luca/ast/c/volatile_qualifier'
require 'luca/ast/c/member_declarator'
# require 'luca/ast/c/struct_type_specifier'
# require 'luca/ast/c/union_type_specifier'
require 'luca/ast/c/storage_class_specifier'
require 'luca/ast/c/typedef_specifier'
require 'luca/ast/c/extern_specifier'
require 'luca/ast/c/static_specifier'
require 'luca/ast/c/auto_specifier'
require 'luca/ast/c/register_specifier'
require 'luca/ast/c/declarator'
require 'luca/ast/c/pointer'
require 'luca/ast/c/direct_declarator'
require 'luca/ast/c/direct_declarator'
require 'luca/ast/c/declarator_name'
require 'luca/ast/c/parentheses_declarator'
require 'luca/ast/c/array_declarator'
require 'luca/ast/c/function_declarator'
require 'luca/ast/c/variable_array_declarator'
require 'luca/ast/c/int_function_declarator'
require 'luca/ast/c/param_decl'
require 'luca/ast/c/decl_specifier'
require 'luca/ast/c/function_specifier'
require 'luca/ast/c/inline_specifier'
require 'luca/ast/c/type_name'
require 'luca/ast/c/initializer'
require 'luca/ast/c/simple_initializer'
require 'luca/ast/c/compound_initializer'
require 'luca/ast/c/initializer_element'
require 'luca/ast/c/designator'
require 'luca/ast/c/subscript_designator'
require 'luca/ast/c/member_designator'
require 'luca/ast/c/compound_literal'
require 'luca/ast/c/init_declarator'
require 'luca/ast/c/decl'
require 'luca/ast/c/function_def'
require 'luca/ast/c/lua_stmt'
require 'luca/ast/c/program'
require 'luca/ast/c/asm'
require 'luca/ast/c/attribute'
require 'luca/ast/c/extension_specifier'
require 'luca/ast/c/declarator_attribute'
require 'luca/ast/c/c_var'
require 'luca/ast/c/lua_var'
require 'luca/ast/c/nonstandard_restrict_qualifier'
require 'luca/ast/c/string_literal'
