module Luca
  class Ast
    class C
      # Cの<tt>_lua</tt>文を表すASTノード
      #
      # @author autopp <autopp.inc@gmail.com>
      #
      class LuaStmt < Stmt
        @is_abstract = false

        # @return [Token]
        attr_reader :lua_tk

        # @return [Token]
        attr_reader :lparen_tk

        # @return [Expr]
        attr_reader :state_expr

        # @return [Token]
        attr_reader :rparen_tk

        # @return [Token]
        attr_reader :lbrace_tk

        # @return [Ast::Lua::Program]
        attr_reader :lua_program
        alias_method :body, :lua_program
        alias_method :program, :lua_program

        # @return [Token]
        attr_reader :rbrace_tk

        # @return [FunctionDef, nil] この<tt>_lua</tt>文を実行するC関数を返す(未設定ならばnil)
        attr_reader :definition

        # @return [String, nil] Clangで解析する際のラッパーC関数名
        attr_accessor :label

        def initialize(lua_tk, lparen_tk, state_expr, rparen_tk,
                       lbrace_tk, lua_program, rbrace_tk)
          expect_var[:lua_tk].of_instance_to_be Token
          expect_var[:lparen_tk].of_instance_to_be Token
          expect_var[:state_expr].of_instance_to_be Expr
          expect_var[:rparen_tk].of_instance_to_be Token
          expect_var[:lbrace_tk].of_instance_to_be Token
          expect_var[:lua_program].of_instance_to_be Lua::Program
          expect_var[:rbrace_tk].of_instance_to_be Token
          super(lua_tk.line, rbrace_tk.line)
        end

        def print_by(cxt)
          cxt << '_lua( ' << state_expr << ' ) {' << Printer.endl
          cxt.next_level do
            cxt.print_ast(lua_program)
          end
          cxt.newline
          cxt << '}'
        end

        def print_c_by(cxt)
          cxt.in_nolmacro_mode do
            cxt.newline
            cxt.write_lmacro(end_pos, end_pos)
          end
        end

        def perform_translate_lua_by(recv, method, *args, override:, &block)
          @lua_program = recv.send(method, lua_program, *args, &block)
          self
        end

        def perform_traverse_lua_by(recv, method, *args, override:, &block)
          recv.send(method, lua_program, *args, override: override, &block)
          nil
        end

        def set_parent_by(cxt)
          cxt.set_parent_of_ast(state_expr, self)
          cxt.set_parent_of_ast(lua_program, self)
        end

        # この<tt>_lua</tt>文を実行するC関数を設定する
        #
        # @note 設定は1回きり
        #
        # @param [FunctionDef] function
        #
        # @return [FunctionDef] function
        #
        def definition=(function)
          fail 'multiple definition setting' if instance_variable_defined?(:@function)

          expect_var[:function].to_be FunctionDef

          @definition = function
        end

        def collect_chunk_by(cxt)
          # 定義元の関数を設定する
          node = self.parent
          loop do
            case node
            when nil
              fail 'definition is not found (THIS IS BUG!)'
            when FunctionDef
              self.definition = node
              node.lua_stmt_list << self
              break
            else
              node = node.parent
            end
          end

          cxt.add_lua_stmt(self)

          # 中身を解析
          cxt.collect_from_ast(lua_program)
        end

        def gen_input_by(cxt)
          cxt.in_nolmacro_mode do
            cxt << "#{label}: " << state_expr << ';' << cxt.endl
            cxt << program
          end
        end

        def gen_code_by(cxt)
          cxt.in_nolmacro_mode do
            cxt << cxt.endl << '{' << cxt.endl

            cxt.next_level do
              loc = cxt.logical_loc_of(lparen_tk.line)
              cxt.write_lmacro(lparen_tk.line, loc.line)
              cxt << "lua_State *#{cxt.state_name} = "
              cxt.in_lmacro_mode do
                cxt << state_expr
              end
              cxt << ';' << cxt.endl
              cxt << program
            end

            cxt << '}'
          end
        end
      end
    end
  end
end
