module Luca
  class Ast
    class Lua
      # Luaの二項算術演算を表すASTノード
      # 
      # 以下が含まれる
      #  - 加算
      #  - 減算
      #  - 乗算
      #  - 除算
      #  - 剰余算
      #  - 塁乗算
      # 
      # 現状の定数畳み込みは数値同士の場合のみ
      #
      # @author autopp <autopp.inc@gmail.com>
      #
      class ArithOpNode < BinOpNode
        @is_abstract = false
        
        def initialize(op_pos, operator, left, right)
          super(op_pos, operator, left, right)
        end
        
        def gen_code_by(cxt)
          # 左辺と右辺の結果をスタックに積む
          cxt << left << cxt.endl << right << cxt.endl
          
          # 計算
          cxt.call_api('arith', operator.api_const)
        end
        
        private_class_method :new
        
        OPERATOR_ADD = OpNode::Operator.new('+', :LUA_OPADD) do |node, left, right|
          if left.is_a?(NumberLiteral) && right.is_a?(NumberLiteral)
            value = left.value + right.value
            if value.is_a?(Integer)
              IntLiteral.new(op_pos, value)
            elsif value.is_a?(Float)
              FloatLiteral.new(op_pos, value)
            else
              fail "undefined type #{value.class}"
            end
          else
            node
          end
        end
        # <tt>+</tt>演算子ノードを生成する
        #
        # @param [Integer] op_pos 演算子の位置
        # @param [Expr] left 左オペランド
        # @param [Expr] right 右オペランド
        #
        # @return [ArithOpNode] 
        # 
        def self.create_add(op_pos, left, right)
          new(op_pos, OPERATOR_ADD, left, right)
        end
        
        OPERATOR_SUB = OpNode::Operator.new('-', :LUA_OPSUB) do |node, left, right|
          if left.is_a?(NumberLiteral) && right.is_a?(NumberLiteral)
            value = left.value - right.value
            if value.is_a?(Integer)
              IntLiteral.new(op_pos, value)
            elsif value.is_a?(Float)
              FloatLiteral.new(op_pos, value)
            else
              fail "undefined type #{value.class}"
            end
          else
            node
          end
        end
        # <tt>-</tt>演算子ノードを生成する
        #
        # @param [Integer] op_pos 演算子の位置
        # @param [Expr] left 左オペランド
        # @param [Expr] right 右オペランド
        #
        # @return [ArithOpNode] 
        # 
        def self.create_sub(op_pos, left, right)
          new(op_pos, OPERATOR_SUB, left, right)
        end
        
        OPERATOR_MUL = OpNode::Operator.new('*', :LUA_OPMUL) do |node, left, right|
          if left.is_a?(NumberLiteral) && right.is_a?(NumberLiteral)
            value = left.value * right.value
            if value.is_a?(Integer)
              IntLiteral.new(op_pos, value)
            elsif value.is_a?(Float)
              FloatLiteral.new(op_pos, value)
            else
              fail "undefined type #{value.class}"
            end
          else
            node
          end
        end
        
        # <tt>*</tt>演算子ノードを生成する
        #
        # @param [Integer] op_pos 演算子の位置
        # @param [Expr] left 左オペランド
        # @param [Expr] right 右オペランド
        #
        # @return [ArithOpNode] 
        # 
        def self.create_mul(op_pos, left, right)
          new(op_pos, OPERATOR_MUL, left, right)
        end
        
        OPERATOR_DIV = OpNode::Operator.new('/', :LUA_OPDIV) do |node, left, right|
          if left.is_a?(NumberLiteral) && right.is_a?(NumberLiteral)
            value = left.value / right.value
            if value.is_a?(Integer)
              IntLiteral.new(op_pos, value)
            elsif value.is_a?(Float)
              FloatLiteral.new(op_pos, value)
            else
              fail "undefined type #{value.class}"
            end
          else
            node
          end
        end
        
        # <tt>/</tt>演算子ノードを生成する
        #
        # @param [Integer] op_pos 演算子の位置
        # @param [Expr] left 左オペランド
        # @param [Expr] right 右オペランド
        #
        # @return [ArithOpNode] 
        # 
        def self.create_div(op_pos, left, right)
          new(op_pos, OPERATOR_DIV, left, right)
        end
        
        OPERATOR_MOD = OpNode::Operator.new('%', :LUA_OPMOD) do |node, left, right|
          if left.is_a?(NumberLiteral) && right.is_a?(NumberLiteral)
            value = left.value % right.value
            if value.is_a?(Integer)
              IntLiteral.new(op_pos, value)
            elsif value.is_a?(Float)
              FloatLiteral.new(op_pos, value)
            else
              fail "undefined type #{value.class}"
            end
          else
            node
          end
        end
        
        # <tt>%</tt>演算子ノードを生成する
        #
        # @param [Integer] op_pos 演算子の位置
        # @param [Expr] left 左オペランド
        # @param [Expr] right 右オペランド
        #
        # @return [ArithOpNode] 
        # 
        def self.create_mod(op_pos, left, right)
          new(op_pos, OPERATOR_MOD, left, right)
        end
        
        OPERATOR_POW = OpNode::Operator.new('^', :LUA_OPPOW) do |node, left, right|
          if left.is_a?(NumberLiteral) && right.is_a?(NumberLiteral)
            value = left.value**right.value
            if value.is_a?(Integer)
              IntLiteral.new(op_pos, value)
            elsif value.is_a?(Float)
              FloatLiteral.new(op_pos, value)
            else
              fail "undefined type #{value.class}"
            end
          else
            node
          end
        end
        
        # <tt>^</tt>演算子ノードを生成する
        #
        # @param [Integer] op_pos 演算子の位置
        # @param [Expr] left 左オペランド
        # @param [Expr] right 右オペランド
        #
        # @return [ArithOpNode] 
        # 
        def self.create_pow(op_pos, left, right)
          new(op_pos, OPERATOR_POW, left, right)
        end
      end
    end
  end
end
