module Luca
  class Ast
    class Lua
      # Luaの代入文を表すASTノード
      #
      # @author autopp <autopp.inc@gmail.com>
      #
      class AssignStmt < Stmt
        @is_abstract = false

        # @return [[Lhs]] 代入の左辺のリスト
        attr_reader :lhs_list

        # @return [[Expr]] 代入の右辺のリスト
        attr_reader :rhs_list

        # @return [Integer] 代入演算子の位置
        attr_reader :op_pos

        def initialize(op_pos, lhs_list, rhs_list)
          expect_var[:op_pos].of_instance_to_be Integer
          expect_var[:lhs_list].to_be ArrayOf(Or(Var, AccessExpr, BackQuoteExpr))
          @lhs_list = lhs_list.map(&:to_lhs)

          expect_var[:rhs_list].of_instance_to_be ArrayOf(Expr)

          super(lhs_list.first.start_pos, rhs_list.last.end_pos)
        end

        def print_by(cxt)
          cxt << Printer.list(lhs_list, ', ') << ' = ' << Printer.list(rhs_list, ', ')
        end

        def desugar_by(cxt)
          @lhs_list = cxt.desugar_list(lhs_list)
          @rhs_list = cxt.desugar_list(rhs_list)

          self
        end

        def set_parent_by(cxt)
          cxt.set_parent_of_list(lhs_list, self)
          cxt.set_parent_of_list(rhs_list, self)
        end

        def collect_chunk_by(cxt)
          cxt.collect_from_list(lhs_list)
          cxt.collect_from_list(rhs_list)
        end

        def analyze_jump_by(cxt)
        end

        def analyze_scope_by(cxt)
          analyze_scope_of_ivar_by(cxt, :@lhs_list)
          analyze_scope_of_ivar_by(cxt, :@rhs_list)
          self
        end

        def analyze_stack_by(cxt)
          cxt.let_safe_int(2) do
            lhs_usage = analyze_stack_of_lhs_by(cxt)
            rhs_usage = analyze_stack_of_rhs_by(cxt)

            # 左辺値計算後のスタックの高さ
            lhs_height = lhs_list.map { |lhs| lhs.stack_usage(cxt) }.reduce(&:+)

            # 代入時に積むオーバーヘッドを計算
            last_overhead = lhs_list.last.assignment_usage(cxt)
            cxt.max(lhs_usage, lhs_height + rhs_usage + last_overhead)
          end
        end

        # 左辺の算出に利用するスタックの高さを計算
        #
        # @param [StackAnalyzer] cxt
        #
        # @return [Integer]
        #
        def analyze_stack_of_lhs_by(cxt)
          lhs_usage_list = lhs_list.map.with_index do |lhs, i|
            lhs.analyze_stack_of_lvalue_by(cxt) + i
          end
          lhs_usage_list.max
        end

        def analyze_stack_of_rhs_by(cxt)
          lhs_size = lhs_list.size
          rhs_size = rhs_list.size
          # 左辺の右辺の個数に応じて分岐
          rhs_usage_list = if lhs_size >= rhs_size
                             # 最後以外は1個積む前提で解析
                             rhs_usage_list = rhs_list[0...-1].map.with_index do |rhs, i|
                               cxt.analyze_ast(rhs, 1) + i
                             end

                             # 最後は左辺の個数に合わせて調節
                             n = lhs_size - rhs_size + 1
                             last_usage = cxt.analyze_ast(rhs_list.last, n)
                             rhs_usage_list << last_usage + rhs_size - 1
                           else
                             # 左辺の個数までは1個ずつ積む前提で解析
                             forward_list = rhs_list.take(lhs_size)
                             rhs_usage_list = forward_list.map.with_index do |rhs, i|
                               cxt.analyze_ast(rhs, 1) + i
                             end

                             # 残りは全て捨てる前提で解析(保守的に計算)
                             backward_list = rhs_list.drop(lhs_size)
                             rhs_usage_list += backward_list.map.with_index do |rhs, i|
                               cxt.analyze_ast(rhs, 0) + i + lhs_size
                             end
                             rhs_usage_list
                           end
          rhs_usage_list.max
        end

        private :analyze_stack_of_rhs_by, :analyze_stack_of_lhs_by

        def gen_input_by(cxt)
          cxt.gen_input_of_list(lhs_list)
          cxt.gen_input_of_list(rhs_list)
        end

        def gen_code_by(cxt)
          cxt.let_safe_int do |before_height_var|
            # 最初のスタックの高さを記憶
            cxt.add_stmt("#{before_height_var} = " + cxt.build_api('gettop'))
            # 左辺値が積むスタックの高さを計算
            lhs_usage_list = lhs_list.map do |lhs|
              lhs.stack_usage(cxt)
            end
            expect_var[:lhs_usage_list].to_be ArrayOf(Integer)

            # 左辺値を計算し，それぞれの戻り値を保存
            # (後でgen_assignment_code_byに渡す)
            state_list = lhs_list.map do |lhs|
              lhs.gen_lvalue_code_by(cxt)
            end

            cxt.let_safe_int do |height_var|
              # この時点でのスタックの高さを保存
              cxt.add_stmt("#{height_var} = #{cxt.build_api('gettop')}")

              # 右辺値を計算し，スタックに積む
              rhs_list.each do |rhs|
                cxt << rhs
              end

              # スタックの高さを右辺値計算前 + 左辺値の個数の高さに調整
              # (すなわち，左辺値の個数に右辺値の個数を調整)
              cxt.call_api('settop', "#{height_var} + #{lhs_list.size}")

              # 右辺値を無視した左辺値のインデックスを計算
              index_list = lhs_usage_list.reverse_each.with_object([0]) do |u, l|
                l << u + l.last
              end

              index_list = index_list.drop(1)
              index_list.reverse!

              # 逆順に代入処理
              lhs_size = lhs_list.size
              lhs_list.zip(state_list, index_list).reverse_each.with_index do |(lhs, state, i), j|
                lhs.gen_assignment_code_by(cxt, state, -(i + (lhs_size - j)))
              end

              # 代入前のスタックの高さに調整(残ってる左辺値を吹っ飛ばす)
              cxt.call_api('settop', before_height_var)
            end
          end
        end
      end
    end
  end
end
