module Luca
  class Ast
    class Lua
      # Luaのバッククォート演算式を表すASTノード
      #
      # @note このクラスは現在は継承を禁止している．
      #   他のASTからバッククォート式への参照問題が未解決のため．
      #
      # @author autopp <autopp.inc@gmail.com>
      #
      class BackQuoteExpr < Expr
        def self.inherited(c)
          fail "#{self} is final class"
        end

        @is_abstract = false

        # @return [Token]
        attr_reader :bquote_tk

        # @return [C::Expr] オペランドとなるCの式
        attr_reader :expr

        # @return [String, nil] Clangで解析する際のラッパーC関数名
        attr_accessor :label

        # @return [FFI::Clang::Type, nil] 解析結果の型
        attr_reader :type

        # @return [String]
        attr_reader :wrapper_function_name

        attr_accessor :at_call_stmt
        def initialize(bquote_tk, expr)
          expect_var[:bquote_tk].of_instance_to_be Token
          expect_var[:expr].of_instance_to_be C::Expr
          @at_call_stmt = false
          super(bquote_tk.line, expr.end_pos)
        end

        def to_lhs
          BackQuoteLhs.new(bquote_tk, expr)
        end

        def type=(type)
          fail 'multiple setting @type' if instance_variable_defined?(:@type)
          expect_var[:type].to_be FFI::Clang::Type
          @type = type.canonical_rec
        end

        def print_by(cxt)
          cxt << bquote_tk << expr
        end

        def desugar_by(cxt)
          self
        end

        def set_parent_by(cxt)
          cxt.set_parent_of_ast(expr, self)
        end

        def collect_chunk_by(cxt)
          # 解析中のC関数にこの逆クォート式を登録
          node = parent
          loop do
            case node
            when nil
              fail 'c function is not found (THIS IS BUG!)'
            when C::FunctionDef
              node.bquote_list << self
              break
            else
              node = node.parent
            end
          end

          cxt.add_bquote(self)
        end

        def analyze_scope_by(cxt)
          analyze_scope_of_ivar_by(cxt, :@expr)
        end

        def gen_input_by(cxt)
          cxt << "#{label}: " << expr << ';' << cxt.endl
        end

        def analyze_stack_by(cxt)
          @expected = cxt.current_expected_result

          if @expected != 0
            if @type.int? || @type.float_point? || @type.string?
              1
            elsif @type.function_pointer?
              2
            elsif @type.void?
              0
            else
              3
            end
          else
            0
          end
        end

        def gen_wrapper_function_by(cxt)
          fail 'should be function' unless type.function_pointer?
          func_type = type.pointee
          @wrapper_function_name = cxt.gen_name(:bquote_function)

          loc = cxt.logical_loc_of(expr.start_pos)
          cxt.add_comment "Wrapper function of backquote expression at #{loc.filename}:#{loc.line}"
          cxt << "int #{wrapper_function_name}(lua_State *#{cxt.state_name})" << cxt.endl

          cxt.nest_block do
            # 引数の個数をチェック
            cxt << "if( #{cxt.build_api('gettop')} != #{func_type.num_arg_types} ) {" << cxt.endl
            cxt.next_level do
              cxt.call_api('pushstring', "\"given #{cxt.build_auxapi('gettop')} for #{func_type.num_arg_types}\"")
            end
            cxt << '}' << cxt.endl

            # 引数を保存する変数名を生成
            arg_var_list = func_type.num_arg_types.times.map do |i|
              cxt.gen_name("arg#{i + 1}")
            end

            # 引数を取得/チェック
            func_type.each_arg_type.with_index do |t, i|
              arg_var = arg_var_list[i]
              stack_idx = i + 1
              cxt.gen_checking_type_code(stack_idx, t)
              decl = t.gen_decl(arg_var)

              api = if t.integer?
                      cxt.build_api('tointeger', stack_idx)
                    elsif t.float_point?
                      cxt.build_api('tonumber', stack_idx)
                    elsif t.string?
                      cxt.build_api('tostring', stack_idx)
                    else
                      "*(#{t.gen_decl('*')})#{cxt.build_api('touserdata', stack_idx)}"
                    end

              cxt.add_stmt("#{decl} = #{api}")
            end

            # 上位値にある関数を取り出す
            func_var = cxt.gen_name('func')
            func_decl = type.gen_decl(func_var)
            rhs = cxt.build_api('touserdata', 'lua_upvalueindex(1)')
            cxt.add_stmt("#{func_decl} = #{rhs}")

            # 関数を呼び出す
            calling_text = "(*#{func_var})(#{arg_var_list.join(', ')})"
            if at_call_stmt
              gen_calling_cfunction_without_result_by(cxt, calling_text)
            else
              gen_calling_cfunction_with_result_by(cxt, calling_text)
            end
          end
        end

        def gen_calling_cfunction_with_result_by(cxt, calling_text)
          result_type = type.pointee.result_type
          logical_line = cxt.logical_loc_of(bquote_tk.line).line

          if result_type.void?
            # 戻り値不要の呼出し
            cxt.add_stmt(calling_text)
          else
            # 戻り値を格納する変数を宣言
            ret_var = cxt.gen_name(:ret)
            cxt.add_stmt(result_type.gen_decl(ret_var))

            # 関数呼出し
            cxt.add_stmt("#{ret_var} = #{calling_text}")
          end

          # 戻り値を変換してスタックにpush
          if result_type.int?
            cxt.call_api('pushinteger', ret_var)
          elsif result_type.float_point?
            cxt.call_api('pushnumber', ret_var)
          elsif result_type.string?
            cxt.call_api('pushstring', ret_var)
          elsif result_type.void?
            cxt.call_api('pushnil')
          else
            udata_var = cxt.gen_name(:udata)
            cxt.gen_userdata_create_code(result_type, udata_var)
            cxt.newline
            cxt.write_lmacro(bquote_tk.line, logical_line)
            cxt << "*#{udata_var} = #{ret_var};" << cxt.endl
          end

          cxt.add_stmt('return 1')
        end

        def gen_calling_cfunction_without_result_by(cxt, calling_text)
          # 呼出し
          cxt.add_stmt(calling_text)
          cxt.add_stmt('return 0')
        end

        def gen_code_by(cxt)
          tmp_var = cxt.gen_name(:tmp)
          logical_line = cxt.logical_loc_of(bquote_tk.line).line
          if @expected == 0
            cxt.newline
            cxt.write_lmacro(bquote_tk.line, logical_line)
            cxt.in_lmacro_mode do
              cxt << expr
            end
            cxt << ';' << cxt.endl
          elsif type.int?
            cxt.newline
            cxt.write_lmacro(bquote_tk.line, logical_line)
            cxt << "int #{tmp_var} = "
            cxt.in_lmacro_mode do
              cxt << expr
            end
            cxt << ';' << cxt.endl
            cxt.call_api('pushinteger', tmp_var)
          elsif type.float_point?
            cxt.newline
            cxt.write_lmacro(bquote_tk.line, logical_line)
            cxt << "double #{tmp_var} = "
            cxt.in_lmacro_mode do
              cxt << expr
            end
            cxt << ';' << cxt.endl
            cxt.call_api('pushnumber', tmp_var)
          elsif type.string?
            cxt.newline
            cxt.write_lmacro(bquote_tk.line, logical_line)
            cxt << "const char *#{tmp_var} = "
            cxt.in_lmacro_mode do
              cxt << expr
            end
            cxt << ';' << cxt.endl
            cxt.call_api('pushstring', tmp_var)
          elsif type.function_pointer?
            fail NotImplementedError, 'not implemented for variadic function' if type.variadic?
            # 関数ポインタをライトユーザデータとして上位値に持つLua関数を生成
            cxt.newline
            cxt.write_lmacro(bquote_tk.line, logical_line)
            cxt << "void *#{tmp_var} = "
            cxt.in_lmacro_mode do
              cxt << expr
            end
            cxt << ';' << cxt.endl
            cxt.call_api('pushlightuserdata', tmp_var)
            cxt.call_api('pushcclosure', wrapper_function_name, 1)
          elsif type.void?
            cxt.newline
            cxt.write_lmacro(bquote_tk.line, logical_line)
            cxt.in_lmacro_mode do
              cxt << expr
            end
            cxt << ';' << cxt.endl
          else
            udata_var = cxt.gen_name(:udata)
            cxt.gen_userdata_create_code(type, udata_var)
            cxt.newline
            cxt.write_lmacro(bquote_tk.line, logical_line)
            cxt << "*#{udata_var} = "
            cxt.in_lmacro_mode do
              cxt << expr
            end
            cxt << ';' << cxt.endl
          end
        end
      end
    end
  end
end
