module Luca
  class Ast
    class Lua
      # Luaのブロックを表すASTノード
      #
      # @author autopp <autopp.inc@gmail.com>
      #
      class Block < Lua
        @is_abstract = false

        # @return [[Stmt]] ブロック本体の文リスト
        attr_reader :stmt_list

        # @return [ReturnStmt, nil] ブロック終端の<tt>return</tt>文
        attr_reader :ret_stmt

        # @return [Block] 外側のブロック
        # @return [FunctionLiteral] 関数内の最も外側のブロックの場合
        # @return [Program] <tt>_lua</tt>文内の最も外側のブロックの場合
        attr_reader :outer_block

        def initialize(stmt_list, ret_stmt = nil)
          @stmt_list = expect_var[:stmt_list].to_be ArrayOf(Stmt)
          @ret_stmt = expect_var[:ret_stmt].to_be Maybe(ReturnStmt)
          @label_list = []

          if stmt_list.empty?
            if ret_stmt
              start_pos = ret_stmt.start_pos
              end_pos = ret_stmt.end_pos
            else
              start_pos = end_pos = 0
            end
          else
            start_pos = stmt_list.first.start_pos
            end_pos = ret_stmt ? ret_stmt.end_pos : stmt_list.last.end_pos
          end
          super(start_pos, end_pos)
        end

        # ブロック先頭の宣言文を返す
        #
        #
        # @return [DeclStmt, nil]
        #
        def decl_stmt
          first = decl_stmt.first
          first && first.decl? ? first : nil
        end

        def ret_stmt=(ret_stmt)
          fail 'multiple setting for @ret_stmt' if @ret_stmt
          expect_var[:ret_stmt].of_instance_to_be ReturnStmt
        end

        def print_by(cxt)
          # 空文以外を改行区切りで出力
          sliced = stmt_list.slice_before { |stmt| !stmt.is_a?(EmptyStmt) }.to_a
          unless sliced.empty?
            cxt << Printer.list(sliced.first, '')
            sliced[1...sliced.length].each do |stmts|
              cxt << Printer.endl << Printer.list(stmts, '')
            end
          end

          cxt << Printer.endl << ret_stmt if ret_stmt
        end

        # ブロックをdesugarする
        #
        # 以下を行う
        #  * ローカル宣言を先頭にのみ許すように変換
        #  * 最後に<tt>return</tt>文があることを保証
        #
        # ローカル関数定義はここで展開する({LocalFunctionDefStmt#desugar_by}は決して呼び出されない)
        #
        # @param [DesugarContext] cxt
        #
        # @return [Ast]
        #
        def desugar_by(cxt)
          # ローカル関数定義の展開
          @stmt_list = stmt_list.each_with_object([]) do |stmt, list|
            if stmt.is_a?(LocalFunctionDefStmt)
              # ローカル関数定義は展開して追加
              # ローカル変数として宣言
              local_decl = LocalDeclStmt.new(stmt.start_pos, stmt.end_pos, [stmt.name], [])

              # 関数リテラルを代入
              lhs = Var.new(stmt.start_pos, stmt.name)
              rhs = FunctionLiteral.new(stmt.start_pos, stmt.end_pos, stmt.param_list, stmt.has_varg?, stmt.body)
              assign_stmt = AssignStmt.new(0, [lhs], [rhs])

              list << local_decl << assign_stmt
            else
              list << stmt
            end
          end

          # ローカル宣言がブロックの先頭のみに現れるように変形
          return self if stmt_list.empty?

          # 先頭のみ無条件に追加
          translated_stmt_list = [stmt_list.first]
          stmt_list[1...stmt_list.length].each_with_index do |stmt, i|
            if stmt.decl?
              # ここでのiは stmt_list[1...stmt_list.length] におけるのi
              # 最初を飛ばしているので+1する必要がある
              rest_block = Block.new(stmt_list[(i + 1)...stmt_list.length], ret_stmt)
              translated_stmt_list << DoStmt.new(rest_block.start_pos, rest_block.end_pos, rest_block)

              # do-endに移動したreturn文を削除
              @ret_stmt = nil
              break
            else
              translated_stmt_list << stmt
            end
          end

          # 個々の文をdesugar
          @stmt_list = translated_stmt_list.map! do |stmt|
            cxt.desugar_ast(stmt)
          end

          @ret_stmt = cxt.desugar_ast(ret_stmt) if ret_stmt
          self
        end

        def set_parent_by(cxt)
          cxt.set_parent_of_list(stmt_list, self)
          cxt.set_parent_of_ast(ret_stmt, self) if ret_stmt
        end

        # 外側のブロックを設定する
        #
        # @param [Block, FunctionLiteral, Program] block
        #
        # @return [Block, FunctionLiteral, Program] block
        #
        def outer_block=(block)
          if instance_variable_defined?(:@outer_block)
            fail 'multiple outer_block setting'
          end

          expect_var[:block].of_instance_to_be Or(Block, FunctionLiteral, Program)
        end

        # このブロック内で宣言されているラベルのリスト(の複製)を返す
        #
        #
        # @return [[LabelStmt]]
        #
        def label_list
          @label_list.dup
        end

        def add_label(label)
          expect_var[:label].to_be LabelStmt

          if @label_list.any? { |l| l.name == label.name }
            fail "label `#{label.name}' already defined in same scope"
          end

          @label_list << label
        end

        def collect_chunk_by(cxt)
          # 外側のブロックを設定
          self.outer_block = cxt.current_block

          cxt.nest_block(self) do
            cxt.collect_from_list(stmt_list)
            cxt.collect_from_ast(ret_stmt) if ret_stmt
          end
        end

        def analyze_jump_by(cxt)
          cxt.analyze_list(stmt_list)
        end

        def analyze_scope_by(cxt)
          unless stmt_list.empty?
            # (宣言文の可能性がある)先頭とそれ以外に分解
            first = stmt_list.first
            rest = stmt_list.drop(1)

            # 先頭を解析
            first = cxt.analyze_ast(first)

            # 宣言文の種類に応じて先頭以外を解析
            case first
            when LocalDeclStmt
              cxt.nest_lua_local_scope(first.var_list) do
                first.var_index_list = first.var_list.map do |var|
                  cxt.lookup_lua_var(var)[1][1]
                end

                first.var_id_list = first.var_list.map do |var|
                  cxt.lookup_lua_var(var)[2]
                end

                rest = cxt.analyze_list(rest)
              end
            when ProtectedDeclStmt
              cxt.nest_protected_scope(first.var_list) do
                rest = cxt.analyze_list(rest)
              end
            when CLocalDeclStmt
              cxt.nest_c_local_scope([first.name]) do
                first.member_index = cxt.lookup_c_var(first.name)[1]
                rest = cxt.analyze_list(rest)
              end
            else
              rest = cxt.analyze_list(rest)
            end

            @stmt_list = [first, *rest]
            expect_var[:@stmt_list].to_be ArrayOf(Stmt)
          end
          analyze_scope_of_ivar_by(cxt, :@ret_stmt) if ret_stmt

          self
        end

        def gen_input_by(cxt)
          cxt << '{' << cxt.endl
          cxt.next_level do
            cxt.gen_input_of_list(stmt_list)
            cxt.gen_input_of_ast(ret_stmt) if ret_stmt
          end
          cxt << '}' << cxt.endl
        end

        def analyze_stack_by(cxt)
          stmt_usage = cxt.analyze_list(stmt_list, 0).max || 0

          ret_usage = 0

          ret_usage = cxt.analyze_ast(ret_stmt) if ret_stmt

          [stmt_usage, ret_usage].max
        end

        def gen_code_by(cxt)
          cxt.nest_block do
            stmt_list.each do |stmt|
              cxt << stmt << cxt.endl
            end
            cxt << ret_stmt << cxt.endl if ret_stmt
          end
        end
      end
    end
  end
end
