module Luca
  class Ast
    class Lua
      # Luaの比較演算を表すASTノード
      # 
      #
      # @author autopp <autopp.inc@gmail.com>
      #
      class CompareOpNode < BinOpNode
        @is_abstract = false
        
        def initialize(op_pos, operator, left, right)
          super(op_pos, operator, left, right)
        end
        
        private_class_method :new
        
        OPERATOR_LTN = OpNode::Operator.new('<', :LUA_OPLT) do |node, left, right|
          if left.number? && right.number? || left.string? && right.string?
            BooleanLiteral.create_constant(node.op_pos, left.value < right.value)
          else
            node
          end
        end
        
        # <tt><</tt>演算子ノードを生成する
        #
        # @param [Integer] op_pos 演算子の位置
        # @param [Expr] left 左オペランド
        # @param [Expr] right 右オペランド
        #
        # @return [CompareOpNode] 
        # 
        def self.create_ltn(op_pos, left, right)
          new(op_pos, OPERATOR_LTN, left, right)
        end
        
        OPERATOR_LEQ = OpNode::Operator.new('<=', :LUA_OPLE) do |node, left, right|
          if left.number? && right.number? || left.string? && right.string?
            BooleanLiteral.create_constant(node.op_pos, left.value <= right.value)
          else
            node
          end
        end

        # <tt><=</tt>演算子ノードを生成する
        #
        # @param [Integer] op_pos 演算子の位置
        # @param [Expr] left 左オペランド
        # @param [Expr] right 右オペランド
        #
        # @return [CompareOpNode] 
        # 
        def self.create_leq(op_pos, left, right)
          new(op_pos, OPERATOR_LEQ, left, right)
        end
        
        OPERATOR_EQL = OpNode::Operator.new('==', :LUA_OPEQ) do |node, left, right|
          if left.number? && right.number? || left.string? && right.string?
            BooleanLiteral.create_constant(node.op_pos, left.value == right.value)
          else
            node
          end
        end
        # <tt>==</tt>演算子ノードを生成する
        #
        # @param [Integer] op_pos 演算子の位置
        # @param [Expr] left 左オペランド
        # @param [Expr] right 右オペランド
        #
        # @return [CompareOpNode] 
        # 
        def self.create_eql(op_pos, left, right)
          new(op_pos, OPERATOR_EQL, left, right)
        end
        
        OPERATOR_GTN = OpNode::Operator.new('>')
        # <tt>></tt>演算子ノードを生成する
        #
        # @param [Integer] op_pos 演算子の位置
        # @param [Expr] left 左オペランド
        # @param [Expr] right 右オペランド
        #
        # @return [CompareOpNode] 
        # 
        def self.create_gtn(op_pos, left, right)
          new(op_pos, OPERATOR_GTN, left, right)
        end
        
        OPERATOR_GEQ = OpNode::Operator.new('<=')
        # <tt>>=</tt>演算子ノードを生成する
        #
        # @param [Integer] op_pos 演算子の位置
        # @param [Expr] left 左オペランド
        # @param [Expr] right 右オペランド
        #
        # @return [CompareOpNode] 
        # 
        def self.create_geq(op_pos, left, right)
          new(op_pos, OPERATOR_GEQ, left, right)
        end
        
        OPERATOR_NEQ = OpNode::Operator.new('~=')
        # <tt>~=</tt>演算子ノードを生成する
        #
        # @param [Integer] op_pos 演算子の位置
        # @param [Expr] left 左オペランド
        # @param [Expr] right 右オペランド
        #
        # @return [CompareOpNode] 
        # 
        def self.create_neq(op_pos, left, right)
          new(op_pos, OPERATOR_NEQ, left, right)
        end
        
        # <tt><</tt>, <tt><=</tt>, <tt>~=</tt>演算子を逆転させた上で定数畳み込みを行う
        # 
        # <tt><</tt>, <tt><=</tt>は左右を反転させる
        # 
        # 
        # @param [DesugarContext] cxt 
        #
        # @return [Expr] desugar後のASTノード
        # 
        def desugar_by(cxt)
          case self.operator
          when OPERATOR_GTN
            cxt.desugar_ast(CompareOpNode.create_ltn(op_pos, right, left))
          when OPERATOR_GEQ
            cxt.desugar_ast(CompareOpNode.create_leq(op_pos, right, left))
          when OPERATOR_NEQ
            cxt.desugar_ast(NotOpNode.new(left.start_pos, CompareOpNode.create_eql(op_pos, left, right)))
          else
            super(cxt)
          end
        end
        
        def gen_code_by(cxt)
          # オペランドをスタックに積むコードを生成
          cxt << left << cxt.endl << right << cxt.endl
          
          # 比較演算APIの呼出し
          tmp = cxt.gen_name('tmp')
          cxt.add_stmt("int #{tmp} = " + cxt.build_api('compare', -2, -1, operator.api_const))
          
          # オペランドをpop
          cxt.call_api('pop', 2)
          
          # 結果をpush
          cxt.call_api('pushboolean', tmp)
        end
      end
    end
  end
end
