module Luca
  class Ast
    class Lua
      # Luaのelseifパートを表すASTノード
      #
      # @author autopp <autopp.inc@gmail.com>
      #
      class ElseifPart < Lua
        @is_abstract = false
        
        # @return [Expr] 条件式
        attr_reader :cond_expr
        
        # @return [Block] 本体
        attr_reader :body
        
        def initialize(start_pos, cond_expr, then_pos, body)
          expect_var[:cond_expr].of_instance_to_be Expr
          expect_var[:body].of_instance_to_be Block
          expect_var[:then_pos].to_be Integer
          
          super(start_pos, body.end_pos)
        end
        
        def print_by(cxt)
          cxt << 'elseif ' << cond_expr << ' then' << Printer.endl
          cxt.next_level do
            cxt.print_ast(body)
          end
        end
        
        def desugar_by(cxt)
          @cond_expr = cxt.desugar_ast(cond_expr)
          @body = cxt.desugar_ast(body)
          
          expect_var[:@cond_expr].to_be Expr
          expect_var[:@body].to_be Block
          
          self
        end
        
        def set_parent_by(cxt)
          cxt.set_parent_of_ast(cond_expr, self)
          cxt.set_parent_of_ast(body, self)
        end
        
        def collect_chunk_by(cxt)
          cxt.collect_from_ast(cond_expr)
          cxt.collect_from_ast(body)
        end
        
        def analyze_jump_by(cxt)
          cxt.analyze_ast(body)
        end
        
        def analyze_scope_by(cxt)
          analyze_scope_of_ivar_by(cxt, :@cond_expr, :@body)
        end
        
        def analyze_stack_by(cxt)
          [
            cxt.expect_result(1) { cxt.analyze_ast(cond_expr) },
            cxt.analyze_ast(body)
          ].max
        end
        
        def gen_input_by(cxt)
          cxt << body
        end
        
        # 
        #
        # @param [CodeGenerator] cxt
        # @param [Symbol] end_label if文終端のラベル
        # @param [current_usage] current_usage 直前までの条件式でpushされている個数
        #
        def gen_code_by(cxt, end_label, current_usage)
          cond_code = if cond_expr.compare_op_node?
                        cxt << cond_expr.left << cxt.endl
                        cxt << cond_expr.right << cxt.endl
                        cxt.build_api('compare', -2, -1, cond_expr.operator.api_const)
                      elsif cond_expr.not_op_node?
                        cxt << cond_expr
                        '!' + cxt.build_api('toboolean', -1)
                      else
                        cxt << cond_expr
                        cxt.build_api('toboolean', -1)
                      end
          gen_body_code_by(cxt, cond_code, end_label, current_usage)
        end
        
        private
        
        def gen_body_code_by(cxt, cond_code, end_label, pop)
          cxt << "if( #{cond_code} )"
          cxt.nest_block do
            # 条件式をpop
            cxt.call_api('pop', pop)
            
            # then節のコードを生成
            cxt << body << cxt.endl
            
            # if文全体の終端へジャンプ
            cxt.add_stmt("goto #{end_label}")
          end
        end
      end
    end
  end
end
