module Luca
  class Ast
    class Lua
      # 汎用<tt>for</tt>文を表すASTノード
      #
      # @author autopp <autopp.inc@gmail.com>
      #
      class ForInStmt < Stmt
        @is_abstract = false

        # @return [[Symbol]] 反復変数のリスト
        attr_reader :var_list

        # @return [[Integer]] 反復対象のインデックスのリスト
        attr_reader :var_index_list

        # @return [[Expr]] 反復対象式のリスト
        attr_reader :expr_list

        # @return [Block] 本体
        attr_reader :body

        def initialize(start_pos, end_pos, var_list, expr_list, body)
          expect_var[:var_list].of_instance_to_be ArrayOf(Symbol)
          expect_var[:expr_list].of_instance_to_be ArrayOf(Expr)
          expect_var[:body].of_instance_to_be Block
          super(start_pos, end_pos)
        end

        def print_by(cxt)
          cxt << 'for ' << Printer.list(var_list) << ' in ' << Printer.list(expr_list) << ' do' << Printer.endl
          cxt.next_level(body)
          cxt << Printer.endl << 'end'
        end

        def desugar_by(cxt)
          desugar_ivar_by(cxt, :@expr_list, :@body)
          expect_var[:expr_list].to_be ArrayOf(Expr)
          expect_var[:body].to_be Block
          self
        end

        def set_parent_by(cxt)
          cxt.set_parent_of_list(expr_list, self)
          cxt.set_parent_of_ast(body, self)
        end

        def collect_chunk_by(cxt)
          cxt.collect_from_list(expr_list)
          cxt.collect_from_ast(body)
        end

        def analyze_jump_by(cxt)
          cxt.nest_loop(self) do
            cxt.analyze_ast(body)
          end
        end

        def analyze_scope_by(cxt)
          analyze_scope_of_ivar_by(cxt, :@expr_list)
          cxt.nest_lua_local_scope(var_list) do
            @var_index_list = var_list.map do |var|
              cxt.lookup_lua_var(var)[1][1]
            end
            @var_id_list = var_list.map do |var|
              cxt.lookup_lua_var(var)[2]
            end

            expect_var[:@var_index_list].to_be ArrayOf(Integer)

            analyze_scope_of_ivar_by(cxt, :@body)
          end

          self
        end

        def analyze_stack_by(cxt)
          # 初期化処理
          init_usage = if expr_list.size <= 3
                         most_usage_list = expr_list[0...- 1].map.with_index do |expr, i|
                           cxt.analyze_ast(expr, 1) + i
                         end
                         most_usage = most_usage_list.max || 0

                         n = [4 - expr_list.size, 0].max
                         last_usage = cxt.analyze_ast(expr_list.last, n) + expr_list.size - 1

                         [most_usage, last_usage].max
                       else
                         head_list = expr_list.take(3)
                         tail_list = expr_list.drop(3)

                         head_usage_list = head_list.map.with_index do |expr, i|
                           cxt.analyze_ast(expr, 1) + i
                         end
                         head_usage = head_usage_list.max || 0

                         tail_usage_list = tail_list.map do |expr|
                           cxt.analyze_ast(expr, 0)
                         end
                         tail_usage = tail_usage_list.max || 0

                         [head_usage, tail_usage + 3].max
                       end

          # 反復処理
          iter_usage = [3, var_list.size].max + 3

          # 本体
          body_usage = cxt.analyze_ast(body) + 3

          [init_usage, iter_usage, body_usage].max
        end

        def gen_input_by(cxt)
          cxt.gen_input_of_list(expr_list)
          cxt.gen_input_of_ast(body)
        end

        def gen_code_by(cxt)
          cxt << expr_list

          if need_settop_call?
            height_var = cxt.gen_name(:height)
            cxt.add_stmt("int #{height_var} = " + cxt.build_api('gettop'))
          end

          # 余分な値をpop
          if expr_list.size < 3
            unless expr_list.last.multi?
              expr_list.size.times do
                cxt.call_api('pushnil')
              end
            end
          elsif need_settop_call?
            cxt.call_api('settop', "#{height_var} + 3")
          end

          cxt << 'while( 1 )'
          cxt.nest_block do
            cxt.call_api('pushvalue', -3)
            cxt.call_api('pushvalue', -3)
            cxt.call_api('pushvalue', -3)
            iter_call_code = cxt.build_api('pcall', 2, var_list.size, 0)

            # iterator関数の呼出し
            cxt << "if( #{iter_call_code} != LUA_OK )"
            cxt.nest_block do
              loc = cxt.logical_loc_of(expr_list.first.start_pos)
              fmt = "\"#{loc.filename}:#{loc.line}: %s\""
              cxt.call_auxapi('error', fmt, cxt.build_api('tostring', -1))
            end

            from_idx = -var_list.size
            to_idx = -var_list.size - 1
            cond_code = cxt.build_api('isnil', from_idx)

            cxt << "if( #{cond_code} )"
            cxt.nest_block do
              cxt.call_api('pop', 3 + var_list.size)
              cxt.add_stmt('break')
            end

            cxt.call_api('copy', from_idx, to_idx)
            var_index_list.zip(@var_id_list).reverse_each do |idx, id|
              cxt.gen_set_local_code(idx, id)
            end

            cxt << body
          end
        end

        def need_settop_call?
          expr_list.size - (expr_list.last.multi? ? 0 : 1) > 3
        end
      end
    end
  end
end
