module Luca
  class Ast
    class Lua
      # Luaの数値用<tt>for</tt>文を表すASTノード
      #
      # @author autopp <autopp.inc@gmail.com>
      #
      class ForStmt < Stmt
        @is_abstract = false

        # @return [Symbol] ループ変数
        attr_reader :var

        # @return [Expr] 初期化式
        attr_reader :init_expr

        # @return [Expr] 上限式
        attr_reader :limit_expr

        # @return [Expr] ステップ式
        attr_reader :step_expr

        # @return [Block] 本体
        attr_reader :body

        def initialize(start_pos, end_pos, var, init_expr, limit_expr, step_expr, body)
          expect_var[:var].of_instance_to_be Symbol
          expect_var[:init_expr].of_instance_to_be Expr
          expect_var[:limit_expr].of_instance_to_be Expr

          expect_var[:step_expr].of_instance_to_be Maybe(Expr)
          expect_var[:body].of_instance_to_be Block

          super(start_pos, end_pos)
        end

        def print_by(cxt)
          cxt << 'for ' << var << ' = ' << init_expr << ', ' << limit_expr
          cxt << ', ' << step_expr if step_expr
          cxt << ' do' << Printer.endl
          cxt.next_level(body)
          cxt << Printer.endl << 'end'
        end

        def desugar_by(cxt)
          desugar_ivar_by(cxt, :@init_expr, :@limit_expr, :@body)
          expect_var[:@init_expr, :@limit_expr].to_be Expr
          expect_var[:@body].to_be Block

          if step_expr
            desugar_ivar_by(cxt, :@step_expr)
            expect_var[:@step_expr].to_be Expr
          end

          self
        end

        def set_parent_by(cxt)
          cxt.set_parent_of_ast(init_expr, self)
          cxt.set_parent_of_ast(limit_expr, self)
          cxt.set_parent_of_ast(step_expr, self) if step_expr
          cxt.set_parent_of_ast(body, self)
        end

        def collect_chunk_by(cxt)
          cxt.collect_from_ast(init_expr)
          cxt.collect_from_ast(limit_expr)
          cxt.collect_from_ast(step_expr) if step_expr
          cxt.collect_from_ast(body)
        end

        def analyze_jump_by(cxt)
          cxt.nest_loop(self) do
            cxt.analyze_ast(body)
          end
        end

        def analyze_scope_by(cxt)
          analyze_scope_of_ivar_by(cxt, :@init_expr, :@limit_expr)
          analyze_scope_of_ivar_by(cxt, :@step_expr) if step_expr

          cxt.nest_lua_local_scope([var]) do
            @var_index = cxt.lookup_lua_var(var)[1][1]
            @var_id = cxt.lookup_lua_var(var)[2]
            analyze_scope_of_ivar_by(cxt, :@body)
          end

          self
        end

        def analyze_stack_by(cxt)
          cxt.let_safe_var(:lua_Number, 3) do
            head_usage = [
              cxt.analyze_ast(init_expr, 1),
              cxt.analyze_ast(limit_expr, 1) + 1,
              step_expr ? cxt.analyze_ast(step_expr, 1) + 2 : 2
            ].max

            body_usage = cxt.max(1, cxt.analyze_ast(body))

            [head_usage, body_usage].max
          end
        end

        def gen_input_by(cxt)
          cxt << init_expr << limit_expr
          cxt << step_expr if step_expr
          cxt << body
        end

        def gen_code_by(cxt)
          cxt.nest_block do
            isnum_var = cxt.gen_name(:isnum)
            cxt.add_stmt("int #{isnum_var}")

            # 初期化式の評価
            cxt.let_safe_var(:lua_Number, 3) do |counter_var, limit_var, step_var|
              cxt << init_expr
              cxt.add_stmt("#{counter_var} = " + cxt.build_api('tonumberx', -1, "&#{isnum_var}"))
              gen_error_check_code_by(cxt, isnum_var, 1, init_expr.start_pos, 'initial value')

              # 上限式の評価
              cxt << limit_expr
              cxt.add_stmt("#{limit_var} = " + cxt.build_api('tonumberx', -1, "&#{isnum_var}"))
              gen_error_check_code_by(cxt, isnum_var, 2, limit_expr.start_pos, 'limit')

              # ステップ式の評価
              if step_expr
                cxt << step_expr
                cxt.add_stmt("#{step_var} = " + cxt.build_api('tonumberx', -1, "&#{isnum_var}"))
                gen_error_check_code_by(cxt, isnum_var, 3, step_expr.start_pos, 'step')
              else
                cxt.add_stmt("#{step_var} = 1")
              end

              cxt.call_api('pop', step_expr ? 3 : 2)

              pos_code = "(#{step_var} > 0 && #{counter_var} <= #{limit_var})"
              neg_code = "(#{step_var} <= 0 && #{counter_var} >= #{limit_var})"
              cmp_code = "#{pos_code} || #{neg_code}"
              cxt << "while( #{cmp_code} )"
              cxt.nest_block do
                # ループ変数に値を格納
                cxt.call_api('pushnumber', counter_var)
                cxt.gen_set_local_code(@var_index, @var_id)

                # 本体のコードを生成
                cxt << body

                # カウンタ更新
                cxt.add_stmt("#{counter_var} += #{step_var}")
              end
            end
          end
        end

        def gen_error_check_code_by(cxt, isnum_var, pop, line, value_kind)
          cxt << "if( !#{isnum_var} )"
          cxt.nest_block do
            cxt.call_api('pop', pop)
            loc = cxt.logical_loc_of(line)
            msg = "\"#{loc.filename}:#{loc.line}: 'for' #{value_kind} must be a number\""
            cxt.call_auxapi('error', msg)
          end
        end
      end
    end
  end
end
