
module Luca
  class Ast
    class Lua
      # Luaの関数リテラルを表すASTノード
      #
      # @author autopp <autopp.inc@gmail.com>
      #
      class FunctionLiteral < Literal
        @is_abstract = false

        # @return [[Symbol]] 仮引数名のリスト
        attr_reader :param_list

        # @return [[Integer]] 仮引数の環境上のインデックスのリスト
        attr_reader :param_index_list

        # @return [Block] 関数本体のブロック
        attr_reader :body

        # @return [FunctionLiteral, Program] 外側の関数もしくはチャンク
        # @return [nil] まだ未設定の場合
        attr_reader :enclosure

        # @return [Block] この関数が定義されているブロック
        attr_reader :defined_block

        # @return [Integer] <description>
        attr_reader :id

        # @return [Hash]
        attr_reader :safe_var_size_map

        # @return [Integer] この関数内で発生する関数呼出しの個数
        attr_accessor :call_count

        def initialize(start_pos, end_pos, param_list, varg, body)
          expect_var[:param_list].of_instance_to_be ArrayOf(Symbol)
          expect_var[:varg].of_instance_to_be Boolean
          expect_var[:body].of_instance_to_be Block

          @inner_function_list = []
          @clocal_decl_list = []

          super(start_pos, end_pos)
        end

        # rubocop:disable Style/TrivialAccessors
        # rubocop:disable Style/PredicateName

        # この関数が可変長引数を持っているか否かを返す
        #
        #
        # @return [Boolean]
        #
        def varg?
          @varg
        end
        # rubocop:enable Style/TrivialAccessors
        # rubocop:enable Style/PredicateName

        def print_by(cxt)
          cxt << 'function (' << Printer.list(param_list, ', ')
          if varg?
            cxt << ', ' unless param_list.empty?
            cxt << '...'
          end
          cxt << ')' << Printer.endl
          cxt.next_level(body)
          cxt << Printer.endl << 'end'
        end

        def desugar_by(cxt)
          desugar_ivar_by(cxt, :@body)

          # 終端にreturn文が無い場合は追加する
          body.ret_stmt = ReturnStmt.new(body.end_pos) unless body.ret_stmt

          self
        end

        def set_parent_by(cxt)
          cxt.set_parent_of_ast(body, self)
        end

        # ネストしている関数を追加する
        #
        # @param [FunctionLiteral] function
        #
        # @return [self]
        #
        def add_inner_function(function)
          expect_var[:function].to_be FunctionLiteral
          @inner_function_list << function
          self
        end

        # ネストしている関数のリスト(のコピー)を返す
        #
        #
        # @return [[FunctionLiteral]]
        #
        def inner_function_list
          @inner_function_list.dup
        end

        # この関数の外側の関数を設定する(or チャンク)
        #
        # @note 設定は1回のみ
        #
        # @param [FunctionLiteral, Program] enc
        #
        # @return [FunctionLiteral, Program] encを返す
        #
        def enclosure=(enclosure)
          if instance_variable_defined?(:@enclosure)
            fail 'multiple enclosure setting'
          end

          expect_var[:enclosure].of_instance_to_be Or(FunctionLiteral, Program)
        end

        # Lua関数のC上の名前を設定する
        #
        # @param [Symbol] name 設定する名前
        #
        # @return [Symbol] <tt>name</tt>をそのまま返す
        #
        def c_name=(name)
          fail 'multiple setting c_name' if instance_variable_defined?(:@c_name)
          expect_var[:name].to_be Symbol

          @c_name = name
        end

        # Lua関数のC上の名前を返す
        #
        #
        # @return [Symbol]
        #
        def c_name
          fail 'no setting c_name' unless instance_variable_defined?(:@c_name)

          @c_name
        end

        # この関数が定義されているブロックを設定する

        def defined_block(block)
          if instance_variable_defined?(:@defined_block)
            fail 'multiple defined_block setting'
          end

          expect_var[:block].of_instance_to_be Block
        end

        def add_c_local_decl(c_local)
          expect_var[:c_local].to_be CLocalDeclStmt
          @clocal_decl_list << c_local
          self
        end

        def clocal_decl_list
          @clocal_decl_list.dup
        end

        def collect_chunk_by(cxt)
          # 現在解析中のチャンクにこの関数を追加する
          cxt.current_chunk.add_inner_function(self)

          # この関数の外側の関数を設定する
          self.enclosure = cxt.current_chunk

          # この関数の本体を解析する
          cxt.nest_chunk(self) do
            cxt.collect_from_ast(body)
          end
        end

        def analyze_jump_by(cxt)
          cxt.analyze_ast(body)
        end

        def analyze_scope_by(cxt)
          cxt.nest_function(self, param_list, varg?) do
            @param_index_list = param_list.map { |param| cxt.lookup_lua_var(param)[1][1] }
            @param_id_list = param_list.map { |param| cxt.lookup_lua_var(param)[2] }
            analyze_scope_of_ivar_by(cxt, :@body)

            # 記憶しておく
            @local_size = cxt.current_function.current_lua_offset
            @id = cxt.current_function.id
            @enclosure = cxt.enclosure_function
            @enclosure_id_list = cxt.current_function_id_list
            @enclosure_id_list.pop(1)
            @enclosure_id_list.reverse!
          end

          self
        end

        def gen_input_by(cxt)
          cxt << body
        end

        def analyze_stack_by(cxt)
          # 本体を解析
          cxt.in_function do
            @body_stack_usage = cxt.analyze_ast(body)
            @safe_var_size_map = cxt.safe_var_size_map
            @call_count = cxt.call_count
          end

          2
        end

        def gen_c_env_by(cxt)
          expect_var[:cxt].to_be CodeGenerator

          env_name = cxt.c_env_name(@id)
          cxt << "struct #{env_name}" << cxt.endl
          cxt.nest_block do
            clocal_decl_list.each do |clocal|
              clocal.gen_env_member_by(cxt)
            end

            # 保護変数を宣言
            @safe_var_size_map.each do |type, size|
              cxt << "#{type} " + cxt.safe_var(type) + "[#{size}];" << cxt.endl
            end
          end
          cxt << ';' << cxt.endl
        end

        # C上の関数定義コードを生成する
        #
        # @param [CodeGenerator] cxt
        #
        # @return [nil]
        #
        def gen_definition(cxt)
          # モードを設定
          cxt.mode = CodeGenerator::MODE_FUNCTION
          cxt.current_function = self
          cxt.current_local_size = @local_size

          # 名前を生成し, 設定
          self.c_name = cxt.gen_name(:func)

          cxt.in_function(id, c_name) do
            cxt << "static int #{c_name}(lua_State *#{cxt.state_name})" << cxt.endl
            cxt.nest_block(true) do
              # スタックサイズを確保
              # 環境の準備の必要な最大サイズ
              env_usage = 1 + 2 * @enclosure_id_list.size

              size = [env_usage, 1 + @body_stack_usage].max
              if varg?
                # 可変長引数の個数を計算
                cxt.add_stmt("int #{cxt.vararg_size} = #{cxt.build_api('gettop')} - #{param_list.size}")
                cxt.add_stmt("#{cxt.vararg_size} = #{cxt.vararg_size} >= 0 ? #{cxt.vararg_size} : 0")

                cxt.call_api('checkstack', "#{size} + #{argc_var}")
              else
                cxt.call_api('checkstack', size)
              end

              # clocal環境へのポインタを宣言
              c_env_type_name = "struct #{cxt.c_env_name(id)}"
              c_env_pointer = cxt.c_env_pointer(id)

              cxt.add_stmt("#{c_env_type_name} *#{c_env_pointer}")

              # 外側のclocal環境へのポインタを全て宣言
              pointer_list = @enclosure_id_list.map { |id| cxt.c_env_pointer(id) }

              @enclosure_id_list.zip(pointer_list).each do |id, pointer|
                type_name = "struct #{cxt.c_env_name(id)} *"
                cxt.add_stmt("#{type_name}#{pointer}")
              end

              # 保護変数を全て宣言
              @safe_var_size_map.each_key do |type|
                cxt.add_stmt("#{type} *#{cxt.safe_var(type)}")
              end

              # この関数がどの状態で呼び出されたかを調べる
              ctx_var = cxt.gen_name(:ctx)
              cxt.add_stmt("int #{ctx_var}")
              cxt << "if( #{cxt.build_api('getctx', "&#{ctx_var}")} == LUA_OK )"
              cxt.nest_block do
                # 普通に呼び出された場合
                # 新しい環境を生成
                cxt.call_api('createtable', 2 + @local_size, 0)

                if enclosure.need_env?
                  # 外側の環境と現在の環境を紐付ける
                  cxt.call_api('pushvalue', 'lua_upvalueindex(1)')
                  cxt.call_api('rawseti', -2, cxt.prev_env_index)
                end

                # C環境の生成
                rhs = cxt.build_api('newuserdata', "sizeof(#{c_env_type_name})")
                cxt.add_stmt("#{c_env_pointer} = #{rhs}")
                cxt.call_api('rawseti', -2, cxt.c_env_index)

                # 外側のclocal環境へのポインタを取り出す
                gen_takeoff_c_env_pointer_code_by(cxt, pointer_list) if enclosure.need_env?

                # 保護変数を取り出す
                gen_takeoff_safe_var_code_by(cxt)

                # 環境をスタックのボトムに挿入
                cxt.call_api('insert', 1)

                # 引数の処理
                if varg?
                  # 引数が足りない場合, nilで補填
                  cxt << "if( #{argc_var} < #{param_list.size} )"
                  cxt.nest_block do
                    cxt.call_api('settop', param_list.size + 1)
                  end

                  cxt << 'else'
                  cxt.nest_block do
                    # 足りている場合は可変長引数をスタックのボトムへ移動
                    var = cxt.gen_name('i')
                    cxt << "for( int #{var} = #{argc_var} - #{param_list.size}; #{var} > 0; #{var}-- )"
                    cxt.nest_block do
                      cxt.call_api('insert', cxt.vararg_offset)
                    end
                  end

                  # 逆順に引数を環境へ格納
                  param_index_list.zip(@param_id_list).reverse_each do |index, id|
                    cxt.gen_set_local_code(index, id)
                  end
                else
                  # 仮引数の数にスタックを調整
                  cxt.call_api('settop', param_list.size + 1)

                  # 逆順に引数を環境へ格納
                  param_index_list.zip(@param_id_list).reverse_each do |index, id|
                    cxt.gen_set_local_code(index, id)
                  end
                end

                # 本体へジャンプ
                cxt.add_stmt("goto #{cxt.body_label}")
              end

              # 以下，継続処理
              # clocal環境へのポインタを取り出す
              cxt.call_api('rawgeti', cxt.env_index, cxt.c_env_index)
              cxt.add_stmt("#{c_env_pointer} = " + cxt.build_api('touserdata', -1))
              cxt.call_api('pop', 1)

              # 外側のclocal環境へのポインタを取り出す
              gen_takeoff_c_env_pointer_code_by(cxt, pointer_list)

              # 保護変数を取り出す
              gen_takeoff_safe_var_code_by(cxt)

              # switch文でgoto
              cxt << "switch( #{ctx_var} )"
              cxt.nest_block do
                1.upto(call_count) do |i|
                  cxt.write_str("case #{i}:", true)
                  cxt.newline
                  cxt.add_stmt("goto #{cxt.after_call_label(i)}")
                  cxt.add_stmt('break')
                end

                cxt.write_str('default:', true)
                cxt.newline
                cxt.call_auxapi('error', "\"invalid ctx %d in #{c_name}\"", ctx_var)
                cxt.add_stmt('break')
              end

              # 関数本体のコードをラベル付きで生成
              cxt.write_str("#{cxt.body_label}:;", true)
              cxt.newline
              cxt << body
            end
          end
        end

        def gen_takeoff_c_env_pointer_code_by(cxt, pointer_list)
          height_var = cxt.gen_name('height')
          cxt.add_stmt("int #{height_var} = " + cxt.build_api('gettop'))
          cxt.call_api('pushvalue', 'lua_upvalueindex(1)')
          pointer_list.each do |pointer|
            cxt.call_api('rawgeti', -1, cxt.c_env_index)
            cxt.add_stmt("#{pointer} = " + cxt.build_api('touserdata', -1))
            cxt.call_api('rawgeti', -2, cxt.prev_env_index)
          end
          cxt.call_api('settop', height_var)
        end

        def gen_takeoff_safe_var_code_by(cxt)
          @safe_var_size_map.each_key do |type|
            lhs = cxt.safe_var(type)
            rhs = "#{cxt.c_env_pointer(id)}->#{cxt.safe_var(type)}"
            cxt.add_stmt("#{lhs} = #{rhs}")
          end
        end

        def gen_code_by(cxt)
          if enclosure.need_env?
            # 現在の環境を上位値に持つクロージャを作成
            cxt.call_api('pushvalue', cxt.env_index)
            cxt.call_api('pushcclosure', self.c_name, 1)
          else
            cxt.call_api('pushcfunction', self.c_name)
          end
        end

        def need_env?
          true
        end
      end
    end
  end
end
