module Luca
  class Ast
    class Lua
      # Luaの<tt>if</tt>文を表すASTノード
      #
      # @author autopp <autopp.inc@gmail.com>
      #
      class IfStmt < Stmt
        @is_abstract = false

        # @return [Expr] 条件式
        attr_reader :cond_expr

        # @return [Block] thenパート
        attr_reader :then_body

        # @return [[ElseifPart]] elseifパートのリスト
        attr_reader :elseif_list

        # @return [Block, nil] elseパート
        attr_reader :else_body

        def initialize(start_pos, end_pos, cond_expr, then_body, elseif_list, else_body = nil)
          expect_var[:cond_expr].of_instance_to_be Expr
          expect_var[:then_body].of_instance_to_be Block
          expect_var[:elseif_list].of_instance_to_be ArrayOf(ElseifPart)
          expect_var[:else_body].of_instance_to_be Maybe(Block)

          super(start_pos, end_pos)
        end

        def print_by(cxt)
          cxt << 'if ' << cond_expr << ' then' << Printer.endl
          cxt.next_level do
            cxt.print_ast(then_body)
          end

          elseif_list.each do |elseif_part|
            cxt << Printer.endl << elseif_part
          end

          if else_part
            cxt << Printer.endl << 'else' << Printer.endl
            cxt.next_level do
              cxt.print_ast(else_body)
            end
          end

          cxt << Printer.endl << 'end'
        end

        def desugar_by(cxt)
          @cond_expr = cxt.desugar_ast(cond_expr)
          @then_body = cxt.desugar_ast(then_body)
          @elseif_list = cxt.desugar_list(elseif_list)
          @else_body = cxt.desugar_ast(else_body) if else_body
          self
        end

        def set_parent_by(cxt)
          cxt.set_parent_of_ast(cond_expr, self)
          cxt.set_parent_of_ast(then_body, self)
          cxt.set_parent_of_list(elseif_list, self)
          cxt.set_parent_of_ast(else_body, self) if else_body
        end

        def collect_chunk_by(cxt)
          cxt.collect_from_ast(cond_expr)
          cxt.collect_from_ast(then_body)
          cxt.collect_from_list(elseif_list)
          cxt.collect_from_ast(else_body) if else_body
        end

        def analyze_jump_by(cxt)
          cxt.analyze_ast(then_body)
          cxt.analyze_list(elseif_list)
          cxt.analyze_ast(else_body) if else_body
        end

        def analyze_scope_by(cxt)
          analyze_scope_of_ivar_by(cxt, :@cond_expr, :@then_body, :@elseif_list)
          analyze_scope_of_ivar_by(cxt, :@else_body) if else_body
          self
        end

        def analyze_stack_by(cxt)
          [
            cxt.analyze_ast(cond_expr, 1),
            cxt.analyze_ast(then_body),
            cxt.analyze_list(elseif_list, 1).max || 0,
            else_body ? cxt.analyze_ast(else_body) : 0
          ].max
        end

        def gen_input_by(cxt)
          cxt.gen_input_of_ast(cond_expr)
          cxt.gen_input_of_ast(then_body)
          cxt.gen_input_of_list(elseif_list)
          cxt.gen_input_of_ast(else_body) if else_body
        end

        def gen_code_by(cxt)
          cxt.add_comment('eval condition of if stmt')

          # 終端のラベルを生成
          end_label = cxt.gen_name('END_OF_IF')

          # ifの条件式を評価
          cond_code = if cond_expr.compare_op_node?
                        cxt << cond_expr.left << cxt.endl
                        cxt << cond_expr.right << cxt.endl
                        cxt.build_api('compare', -2, -1, cond_expr.operator.api_const)
                      elsif cond_expr.not_op_node?
                        cxt << cond_expr
                        '!' + cxt.build_api('toboolean', -1)
                      else
                        cxt << cond_expr
                        cxt.build_api('toboolean', -1)
                      end
          # then節のコードを生成
          gen_then_part_code_by(cxt, cond_code, end_label, cond_stack_usage_list.first)

          # elseifパートのコードを生成
          elseif_list.zip(cond_scan_usage_list).each do |elseif_part, usage|
            cxt.gen_code_of_ast(elseif_part, end_label, usage)
          end

          # これまでpushされた条件式をpop
          cxt.call_api('pop', cond_stack_usage_list.reduce(&:+))
          # else部のコードを生成
          cxt << else_body if else_body

          cxt.newline
          # 終端ラベルの生成
          cxt.add_stmt("#{end_label}:")
        end

        private

        # if, elseif節の条件式で積まれるスタックの高さを計算
        #
        #
        # @return [[Int]]
        #
        def cond_stack_usage_list
          elseif_list.each_with_object([cond_stack_usage(cond_expr)]) do |elseif_part, ls|
            ls << cond_stack_usage(elseif_part.cond_expr)
          end
        end

        def cond_scan_usage_list
          ret = elseif_list.each_with_object([cond_stack_usage(cond_expr)]) do |elseif_part, ls|
            ls << (cond_stack_usage(elseif_part.cond_expr) + ls.last)
          end

          ret.drop(1)
        end

        def cond_stack_usage(expr)
          expr.compare_op_node? ? 2 : 1
        end

        def gen_then_part_code_by(cxt, cond_code, end_label, pop)
          cxt << "if( #{cond_code} )"
          cxt.nest_block do
            # 条件式をpop
            cxt.call_api('pop', pop)

            # then節のコードを生成
            cxt << then_body << cxt.endl

            # if文全体の終端へジャンプ
            cxt.add_stmt("goto #{end_label}")
          end
        end
      end
    end
  end
end
