module Luca
  class Ast
    class Lua
      # Luaの<tt>local</tt>宣言を表すASTノード
      #
      # @author autopp <autopp.inc@gmail.com>
      #
      class LocalDeclStmt < DeclStmt
        @is_abstract = false

        # @return [[Symbol]] 宣言する変数名のリスト
        attr_reader :var_list

        # @return [[Expr]] 初期化式のリスト
        attr_reader :expr_list

        # @return [Integer] 代入演算子の位置(初期化式が無い場合は0)
        attr_reader :op_pos

        # @return [[Integer]] 宣言する変数の環境上のインデックスのリスト
        attr_reader :var_index_list

        # @return [[Object]] 宣言する変数のID
        attr_reader :var_id_list

        def initialize(start_pos, op_end_pos, var_list, expr_list)
          expect_var[:var_list].of_instance_to_be ArrayOf(Symbol) >= 1
          expect_var[:expr_list].of_instance_to_be ArrayOf(Expr)
          expect_var[:op_end_pos].to_be Integer
          @op_pos = expr_list.empty? ? 0 : op_end_pos
          end_pos = expr_list.empty? ? op_end_pos : expr_list.last.end_pos
          super(start_pos, end_pos)
        end

        def print_by(cxt)
          cxt << 'local ' << Printer.list(var_list)

          cxt << ' = ' << Printer.list(expr_list) unless expr_list.empty?
        end

        def desugar_by(cxt)
          @expr_list = cxt.desugar_list(expr_list)
          self
        end

        def set_parent_by(cxt)
          cxt.set_parent_of_list(expr_list, self)
        end

        def collect_chunk_by(cxt)
          cxt.collect_from_list(expr_list)
        end

        def analyze_jump_by(cxt)
        end

        def analyze_scope_by(cxt)
          analyze_scope_of_ivar_by(cxt, :@expr_list)
        end

        def var_index_list=(index_list)
          expect_var[:index_list].to_be ArrayOf(Integer)
          @var_index_list = index_list
        end

        def var_id_list=(id_list)
          expect_var[:id_list].to_be Array
          @var_id_list = id_list
        end

        def var_index_list
          fail "no setting @var_index_list for #{inspect}" unless instance_variable_defined?(:@var_index_list)
          @var_index_list
        end

        def analyze_stack_by(cxt)
          cxt.let_safe_int(1) do
            lhs_size, rhs_size = var_list.size, expr_list.size
            usage_list = if rhs_size == 0
                           [0]
                         elsif lhs_size >= rhs_size
                           # 最後以外は1つ積む前提で解析
                           usage_list = expr_list[0...-1].map.with_index do |rhs, i|
                             cxt.analyze_ast(rhs, 1) + i
                           end

                           # 最後は左辺の個数に合わせて調整
                           n = lhs_size - rhs_size + 1
                           last_usage = cxt.analyze_ast(expr_list.last, n)
                           usage_list << last_usage + rhs_size - 1
                         else
                           # 左辺の個数までは1個ずつ積む前提で解析
                           forward_list = expr_list.take(lhs_size)
                           usage_list = forward_list.map.with_index do |rhs, i|
                             cxt.analyze_ast(rhs, i)
                           end
                           # 残りは全て捨てる前提で解析
                           backward_list = expr_list.drop(lhs_size)
                           usage_list + backward_list.map_with_index do |rhs, i|
                             cxt.analyze_ast(rhs, 0) + i + lhs_size
                           end
                         end
            usage_list.max
          end
        end

        def multi_rhs?
          expr_list.size > 0 && expr_last.multi?
        end

        def gen_input_by(cxt)
          cxt.gen_input_of_list(expr_list)
        end

        def gen_code_by(cxt)
          cxt.let_safe_int do |height_var|
            # スタックの高さを確認
            if need_settop_call?
              cxt.add_stmt("#{height_var} = " + cxt.build_api('gettop'))
            end

            # 右辺値を全てスタックに積む
            expr_list.each do |rhs|
              cxt << rhs
            end

            # 左辺の数に合わせてスタックの高さを調整
            cxt.call_api('settop', "#{height_var} + #{var_list.size}") if need_settop_call?

            # 逆順に格納
            var_index_list.zip(var_id_list).reverse_each do |index, id|
              cxt.gen_set_local_code(index, id)
            end
          end
        end

        # 代入処理前にsettopによるスタックの調整が必要かどうか
        #
        #
        # @return [Boolean]
        #
        def need_settop_call?
          lhs_size, rhs_size = var_list.size, expr_list.size
          if lhs_size > rhs_size
            !multi_rhs?
          elsif lhs_size == rhs_size
            false
          else
            expr_list.drop(lhs_size).all?(&:multi?)
          end
        end
      end
    end
  end
end
