module Luca
  class Ast
    class Lua
      # Luaのメソッド形式の関数呼出し式を表すASTノード
      #
      # @author autopp <autopp.inc@gmail.com>
      #
      class MethodCallExpr < Expr
        @is_abstract = false
        
        # @return [Expr] レシーバ式
        attr_reader :recv_expr
        alias_method :recv, :recv_expr
        
        # @return [Symbol] メソッド名
        attr_reader :method_name
        alias_method :method, :method_name
        
        # @return [[Expr]] 実引数のリスト
        attr_reader :arg_list
        
        def initialize(end_pos, recv_expr, method_name, arg_list)
          expect_var[:recv_expr].of_instance_to_be Expr
          expect_var[:method_name].of_instance_to_be Symbol
          expect_var[:arg_list].of_instance_to_be ArrayOf(Expr)
          
          super(recv_expr.start_pos, end_pos)
        end
        
        def multi?
          true
        end
        
        # 最後の引数が多値か否かを返す
        #
        #
        # @return [Boolean]
        # 
        def multi_last_arg?
          !arg_list.empty && arg_list.last.multi?
        end
        
        def print_by(cxt)
          cxt << recv_expr << ":#{method_name}(" << Printer.list(arg_list, ', ') << ')'
        end
        
        def desugar_by(cxt)
          desugar_ivar_by(cxt, :@recv_expr, :@arg_list)
          self
        end
        
        def set_parent_by(cxt)
          cxt.set_parent_of_ast(recv_expr, self)
          cxt.set_parent_of_list(arg_list, self)
        end
        
        def collect_chunk_by(cxt)
          cxt.collect_from_ast(recv_expr)
          cxt.collect_from_list(arg_list)
        end
        
        def analyze_scope_by(cxt)
          analyze_scope_of_ivar_by(cxt, :@recv_expr, :@arg_list)
        end
        
        def analyze_stack_by(cxt)
          cxt.inc_call_count
          cxt.let_safe_int(1) do
            @expected = cxt.current_expected_result
            @expected = 'LUA_MULTRET' if @expected == cxt.multi_ret
            
            func_usage = cxt.analyze_ast(recv_expr, 1) + 1
            
            arg_usage_list = arg_list[0...-1].map.with_index do |arg, i|
              cxt.analyze_ast(arg, 1) + i + 1
            end
            arg_usage = arg_usage_list.max || 0
            
            last_usage = if arg_list.empty?
                           0
                         else
                           cxt.analyze_ast(arg_list.last, cxt.multi_ret) + arg_list.size - 1
                         end
            
            arg_usage = last_usage if last_usage > arg_usage
            
            cxt.max(func_usage, 2 + arg_usage)
          end
        end
        
        def gen_input_by(cxt)
          cxt.gen_input_of_ast(recv_expr)
          cxt.gen_input_of_list(arg_list)
        end
        
        def gen_code_by(cxt)
          # レシーバをpush
          cxt << recv_expr
          
          # メソッドを取り出し, トップを入れ替える
          cxt.call_api('getfield', -1, "\"#{method_name}\"")
          cxt.call_api('insert', -2)
          
          cxt.let_safe_int do |before_height_var|
            if multi_last_arg?
              cxt.add_stmt("#{before_height_var} = " + cxt.build_api('gettop'))
            end
            
            # 引数を積む
            arg_list.each do |arg|
              cxt << arg
            end
            
            arg_size = if multi_last_arg?
                         cxt.build_api('gettop') + " - #{before_height_var}"
                       else
                         arg_list.size
                       end
            
            # メソッドを呼び出す
            if cxt.mode == CodeGenerator::MODE_FUNCTION
              count = cxt.next_call_count
              cxt.call_api('callk', arg_size + 1, @expected, count, cxt.function_name)
              
              # ラベル生成
              label = cxt.after_call_label(count)
              cxt.write_str("#{label}:;", true)
              cxt.newline
            else
              cxt.call_api('callk', arg_size + 1, @expected, 0, 0)
            end
          end
        end
      end
    end
  end
end
