module Luca
  class Ast
    class Lua
      # 演算子式を表すASTノード
      # @note このクラスは抽象クラスである
      # @author autopp <autopp.inc@gmail.com>
      #
      class OpNode < Expr
        @is_abstract = true
        
        # Luaの演算子を表すクラス
        # 
        # プリティプリント時のテキスト表現と, API関数の為の定数を保持する
        # 
        # @note このクラスは抽象クラスである
        # @author autopp <autopp.inc@gmail.com>
        #
        class Operator
          include Varspec::BuiltinMatcher
          
          # @return [String] 演算子の文字列表現
          attr_reader :text
          
          # @return [Symbol, nil] API関数用のシンボル(必要ない場合は<tt>nil</tt>)
          attr_reader :api_const
          
          def initialize(text, api_const = nil, &const_folding_block)
            expect_var[:text].of_instance_to_be String
            expect_var[:api_const].of_instance_to_be Maybe(Symbol)
            expect_var[:const_folding_block].of_instance_to_be Maybe(Proc)
          end
          
          # 初期化時に与えられたブロックを用いて, 定数畳み込みを行う
          #
          # @param [[Constant]] operand_list オペランドの配列
          #
          # @return [Constant] 畳み込んだ結果
          # 
          def fold_constant(operand_list)
            const_folding_block.call(operand_list)
          end
        end

        # @return [Integer] 演算子の位置
        attr_reader :op_pos
        
        # @return [Operator] 演算子
        attr_reader :operator
        
        # @return [[Expr]] オペランドのリスト
        attr_reader :operand_list
        
        def initialize(start_pos, end_pos, op_pos, operator, operand_list)
          fail "#{self.class} is abstract class" if self.class == OpNode
          
          expect_var[:op_pos].of_instance_to_be Integer
          expect_var[:operator].of_instance_to_be Operator
          expect_var[:operand_list].of_instance_to_be ArrayOf(Expr)
          
          super(start_pos, end_pos)
        end
        
        # <tt>cxt</tt>を用いて<tt>self</tt>をdesugarする
        # 
        # オペランドを全てdesugarし, それらが全て定数ならば畳み込む
        #
        # @param [DesugarContext] cxt 
        #
        # @return [OpNode] 
        # @return [Constant] 定数畳み込みが行われた場合
        # 
        def desugar_by(cxt)
          # オペランドをdesugarする
          desugar_ivar_by(cxt, :@operand_list)
          
          # オペランドが全て定数ならば畳み込む
          if operand_list.all?(&:constant?)
            operator.fold_constant(self, *operand_list)
          else
            self
          end
        end
        
        def set_parent_by(cxt)
          cxt.set_parent_of_list(operand_list, self)
        end
        
        def collect_chunk_by(cxt)
          cxt.collect_from_list(operand_list)
        end
        
        def analyze_scope_by(cxt)
          analyze_scope_of_ivar_by(cxt, :@operand_list)
        end
        
        def analyze_stack_by(cxt)
          cxt.analyze_list(operand_list, 1).map.with_index { |usage, i| usage + i }.max
        end
        
        def gen_input_by(cxt)
          cxt.gen_input_of_list(operand_list)
        end
      end
    end
  end
end
