module Luca
  class Ast
    class Lua
      # Luaプログラム全体を表すASTノード
      #
      # @author autopp <autopp.inc@gmail.com>
      #
      class Program < Lua
        @is_abstract = false

        # @return [Block]
        attr_reader :block

        # @return [Integer] Luaのローカル変数の数を返す
        attr_reader :lua_local_size

        # @return [Integer] <description>
        attr_reader :id

        # @return [Hash]
        attr_reader :safe_var_size_map

        def initialize(block)
          expect_var[:block].of_instance_to_be Block
          @inner_function_list = []
          @clocal_decl_list = []
          super(block.start_pos, block.end_pos)
        end

        def print_by(cxt)
          cxt.print_ast(block)
        end

        def desugar_by(cxt)
          @block = cxt.desugar_ast(block)
          self
        end

        def set_parent_by(cxt)
          cxt.set_parent_of_ast(block, self)
        end

        # ネストしている関数を追加する
        #
        # @param [FunctionLiteral] function
        #
        # @return [self]
        #
        def add_inner_function(function)
          expect_var[:function].to_be FunctionLiteral
          @inner_function_list << function
          self
        end

        # ネストしている関数のリスト(のコピー)を返す
        #
        #
        # @return [[FunctionLiteral]]
        #
        def inner_function_list
          @inner_function_list.dup
        end

        def add_c_local_decl(clocal)
          expect_var[:clocal].to_be CLocalDeclStmt
          @clocal_decl_list << clocal
          self
        end

        def clocal_decl_list
          @clocal_decl_list.dup
        end

        def collect_chunk_by(cxt)
          cxt.nest_chunk(self) do
            cxt.collect_from_ast(block)
          end
        end

        def analyze_jump_by(cxt)
          cxt.analyze_ast(block)

          # 内部のLua関数は個別に解析
          inner_function_list.each do |function|
            cxt.analyze_ast(function)
          end
        end

        def analyze_scope_by(cxt)
          cxt.nest_function(self, [], false) do
            analyze_scope_of_ivar_by(cxt, :@block)

            # ローカル変数の数とIDを記憶しておく
            @lua_local_size = cxt.current_function.current_lua_offset
            @escape_var_count = cxt.escape_var_count
            @id = cxt.current_function.id
          end

          self
        end

        def analyze_stack_by(cxt)
          cxt.in_function do
            @body_stack_usage = cxt.analyze_ast(block)
            @safe_var_size_map = cxt.safe_var_size_map
          end
          @body_stack_usage
        end

        def gen_input_by(cxt)
          cxt.gen_input_of_ast(block)
        end

        def gen_c_env_by(cxt)
          expect_var[:cxt].to_be CodeGenerator

          env_name = cxt.c_env_name(@id)
          cxt << "struct #{env_name}" << cxt.endl
          cxt.nest_block do
            clocal_decl_list.each do |clocal|
              clocal.gen_env_member_by(cxt)
            end
            # 保護変数を宣言
            @safe_var_size_map.each do |type, size|
              cxt << "#{type} " + cxt.safe_var(type) + "[#{size}];" << cxt.endl
            end
          end
          cxt << ';' << cxt.endl
        end

        def gen_code_by(cxt)
          cxt.mode = CodeGenerator::MODE_CHUNK
          cxt.current_local_size = lua_local_size
          cxt.current_function = self

          cxt.end_of_lua_stmt = cxt.gen_name(:END_OF_LUA)

          # 環境を構築
          cxt.call_api('createtable', 1 + lua_local_size, 0) if need_env?

          # ベースのスタックの高さを保存
          cxt.add_stmt("int #{cxt.base_index} = #{cxt.build_api('gettop')}")

          # スタックサイズを確保
          cxt.call_api('checkstack', "#{@body_stack_usage} + #{cxt.base_index}")

          # ローカル変数の分を確保
          cxt.call_api('settop', "#{cxt.base_index} + #{lua_local_size}")

          cxt.nest_block(true) do
            if need_env?
              # C環境を構築
              env_type_name = "struct #{cxt.c_env_name(@id)}"
              env_pointer = cxt.c_env_pointer(@id)
              rhs = cxt.build_api('newuserdata', "sizeof(#{env_type_name})")
              cxt.add_comment('Create environment for clocal variables of chunk')
              cxt.add_stmt("#{env_type_name} *#{env_pointer} = #{rhs}")
              cxt.call_api('rawseti', cxt.env_index, cxt.c_env_index)
            end

            # 保護変数を全て宣言
            if need_env?
              @safe_var_size_map.each_key do |type|
                lhs = "#{type} *#{cxt.safe_var(type)}"
                rhs = "#{env_pointer}->#{cxt.safe_var(type)}"
                cxt.add_stmt("#{lhs} = #{rhs}")
              end
            else
              @safe_var_size_map.each do |type, size|
                cxt.add_stmt("#{type} #{cxt.safe_var(type)}[#{size}]")
              end
            end

            cxt << block
          end

          # _lua文終端のラベルを生成
          cxt.write_str("#{cxt.end_of_lua_stmt}:;", true)
          cxt.newline
          # 環境とスタック上のゴミを削除
          # return で抜けた場合、ゴミが残る
          cxt.call_api('settop', "#{cxt.base_index}#{need_env? ? '-1' : ''}")
          cxt.end_of_lua_stmt = nil
        end

        def need_env?
          !clocal_decl_list.empty? || @escape_var_count > 0
        end
      end
    end
  end
end
