module Luca
  # コード生成器を表すクラス.
  # 各ASTノードクラスはメソッド{get_code_by}を実装する必要がある.
  #
  # @author autopp <autopp.inc@gmail.com>
  #
  class CodeGenerator < Printer
    include Varspec::BuiltinMatcher

    # @return [Compiler]
    attr_reader :compiler

    # @return [Symbol] 生成するコードでの<tt>lua_State</tt>の変数名
    attr_reader :state_name

    # @return [Symbol] 生成するコードでの可変長引数の長さを格納する変数名
    attr_reader :vararg_size

    # @return [Symbol] 生成するコードでの_lua文から脱出する際の戻り値の数の変数名
    attr_reader :ret_size

    # @return [Symbol] 現在の<tt>_lua</tt>文の終端を表すラベル
    attr_accessor :end_of_lua_stmt

    attr_accessor :member_index

    attr_accessor :current_function

    # @return [Symbol] 現在の<tt>_lua</tt>文のベースのスタックの高さを保存する変数名
    attr_reader :base_index

    # @return [Symbol] 関数本体のコードのラベル
    attr_reader :body_label

    attr_reader :function_id
    attr_reader :function_name

    # @return [Integer] 現在のローカル変数の個数
    attr_accessor :current_local_size

    def initialize(compiler, filename, output, width, loc_map, state_name: :LuCa_L)
      super(filename, output, width, loc_map)
      @compiler = compiler
      @name_table = {}
      @safe_var_map = {}
      expect_var[:state_name].of_instance_to_be Symbol
      @vararg_size = :LuCa_vararg_size
      @ret_size = :LuCa_ret_size
      @base_index = :LuCa_base_index
      @body_label = :LuCa_BODY
    end

    def gen_code(program)
      expect_var[:program].of_instance_to_be Ast::C::Program
      gen_code_of_ast(program)
      compiler.flush_error
    end

    # {Printer#print_ast}と異なり, ASTノードのメソッド<tt>gen_code_by</tt>を呼び出す.
    #
    # @param [Ast] ast 出力するASTノード
    #
    # @return [self]
    #
    def print_ast(ast, *args)
      expect_var[:ast].to_be Ast

      ast.gen_code_by(self, *args)
      self
    end

    alias_method :gen_code_of_ast, :print_ast
    alias_method :gen_code_of_list, :print_ast_list

    def gen_name(kind)
      expect_var[:kind].to_be Or(String, Symbol)
      kind = kind.to_sym

      if @name_table.member?(kind)
        v = @name_table[kind] + 1
      else
        v = 0
      end
      @name_table[kind] = v

      "LuCa_#{kind}_#{v}".to_sym
    end

    C_ENV_PREFIX = 'LuCa_cenv'
    def c_env_name(id)
      "#{C_ENV_PREFIX}#{id}".to_sym
    end

    ENV_MEMBER_PREFIX = 'LuCa_env_member'
    def env_member_name(index)
      "#{ENV_MEMBER_PREFIX}#{index}".to_sym
    end

    C_ENV_POINTER_PREFIX = 'LuCa_cenv_pointer'
    def c_env_pointer(id)
      "#{C_ENV_POINTER_PREFIX}#{id}".to_sym
    end

    # Cのブロックをネストする
    #
    #
    # @return [self]
    #
    def nest_block(newline = true, &block)
      self << '{' << endl
      next_level(&block)
      self.newline unless @is_newline
      self << '}'
      self << endl if newline
    end

    # Cの文を追加する.

    #
    # セミコロンと改行を自動的に付与する
    #
    # @param [Symbol, String] str 追加するCの文(セミコロン省略可)
    #
    # @return [self]
    #
    def add_stmt(str)
      expect_var[:str].to_be Or(Symbol, String)

      self << "#{str};" << endl
    end

    # Cのコメントを追加する
    #
    # @param [String] str
    #
    # @return [self]
    #
    def add_comment(str)
      str.each_line do |line|
        self << "// #{line.chomp}" << endl
      end
    end

    def call_api(name, *args, prefix: 'lua_')
      str = build_api(name, *args, prefix: prefix)
      add_stmt(str)
    end

    def call_auxapi(name, *args)
      call_api(name, *args, prefix: 'luaL_')
    end

    def build_api(name, *args, prefix: 'lua_')
      arg_str = args.map { |arg| ", #{arg}" }.join
      "#{prefix}#{name}(#{state_name}#{arg_str})"
    end

    def build_auxapi(name, *args)
      build_api(name, *args, prefix: 'luaL_')
    end

    PREV_ENV_INDEX = 1
    # @return [Integer] 環境内の上位クロージャを格納するインデックス
    def prev_env_index
      PREV_ENV_INDEX
    end

    C_ENV_INDEX = PREV_ENV_INDEX + 1
    def c_env_index
      C_ENV_INDEX
    end

    LUA_LOCAL_OFFSET = C_ENV_INDEX + 1
    # @return [Integer] 環境内のローカル変数を格納するインデックスのオフセット
    def lua_local_offset
      LUA_LOCAL_OFFSET
    end

    def stack_local_offset
      "#{env_index} + 1"
    end

    # 現在生成中のチャンク/関数の環境を格納しているスタック上のインデックスを返す.
    #
    # チャンクの場合は_lua文開始時のインデックス, 関数の場合は1(ボトム)
    #
    # @return [Intger, Symbol]
    #
    def env_index
      case mode
      when MODE_FUNCTION
        1
      when MODE_CHUNK
        base_index
      end
    end

    def vararg_offset
      "#{env_index}  + 1"
    end

    def endl
      Printer.endl
    end

    def list(*args)
      Printer.list(*args)
    end

    USERDATA_TYPE_INDEX = 1
    # ユーザデータ中のuservalueに格納する型情報のインデックスを返す
    #
    #
    # @return [Integer]
    #
    def userdata_type_index
      USERDATA_TYPE_INDEX
    end

    MODE_FUNCTION = :function

    MODE_CHUNK = :chunk
    def mode=(mode)
      fail 'invalid mode' unless [MODE_FUNCTION, MODE_CHUNK].find(mode)

      @mode = mode
    end

    def mode
      fail 'invalid mode' unless [MODE_FUNCTION, MODE_CHUNK].find(@mode)

      @mode
    end

    def warn(msg)
      $stderr.puts "warning: #{msg}"
    end

    def in_function(id, name)
      prev_id, prev_name = @id, @name
      @function_id, @function_name = id, name
      @call_count = 0
      ret = yield
      @function_id, @function_name = prev_id, prev_name
      ret
    end

    def next_call_count
      @call_count += 1
    end

    def after_call_label(i)
      "LuCa_AFTER_CALL#{i}"
    end

    def gen_set_local_code(index, id)
      if escape?(id) || current_function.need_env?
        add_comment 'set escape var'
        call_api('rawseti', env_index, lua_local_offset + index)
      else
        add_comment 'set non escape var'
        call_api('copy', -1, "#{stack_local_offset} + #{index}")
        call_api('pop', 1)
      end
    end

    # type型のユーザデータを生成してスタックトップに置くコードを生成する
    #
    # @param [FFI::Clang::Type] type
    #
    # @return [self]
    #
    def gen_userdata_create_code(type, name)
      # ユーザデータ生成
      decl = type.gen_decl("*#{name}")
      rhs = build_api('newuserdata', type.sizeof)
      add_stmt("#{decl} = #{rhs}")

      # メタ情報を格納するuservalueを生成
      call_api('createtable', 1, 0)
      # 型情報を格納
      call_api('pushstring', "\"#{type.spelling}\"")
      call_api('rawseti', -2, userdata_type_index)

      # uservalueをセット
      call_api('setuservalue', -2)
      self
    end

    # 保護変数周りのインタフェース
    # 保護変数リストのメンバ名
    SAFE_VAR_PREFIX = :LuCa_safe_var_
    def safe_var_prefix
      SAFE_VAR_PREFIX
    end

    def safe_var(type)
      "#{SAFE_VAR_PREFIX}#{type}"
    end

    def let_safe_var(type, n = 1)
      type = type.to_sym
      @safe_var_map[type] = 0 unless @safe_var_map.member?(type)
      prev = @safe_var_map[type]
      @safe_var_map[type] += n
      args = n.times.map { |i| "#{safe_var(type)}[#{prev + i}]" }
      if n == 1
        ret = yield args.first
      else
        ret = yield args
      end
      @safe_var_map[type] = prev
      ret
    end

    def let_safe_int(n = 1, &block)
      let_safe_var(:int, n, &block)
    end

    def escape?(id)
      @program.escape?(id)
    end

    def gen_checking_type_code(index, type, pos_str = nil)
      if type.number?
        gen_checking_number_code(index, type)
      elsif type.string?
        gen_checking_string_code(index, type)
      else
        gen_checking_udata_code(index, type)
      end
    end

    private

    def gen_checking_number_code(index, type)
      # 数値型か調べるコードを生成
      self << "if( #{build_api('type', index)} != LUA_TNUMBER ){" << endl
      next_level do
        call_auxapi('error', "\"expected number, but got %s\"", build_api('typename', build_api('type', index)))
      end
      self << '}' << endl
    end

    def gen_checking_string_code(index, type)
      # 数値型か調べるコードを生成
      self << "if( #{build_api('type', index)} != LUA_TSTRING ){" << endl
      next_level do
        call_auxapi('error', "\"expected string, but got %s\"", build_api('typename', build_api('type', index)))
      end
      self << '}' << endl
    end

    def gen_checking_udata_code(index, type)
      if type.pointer?
        gen_checking_pointer_udata_code(index, type)
      else
        gen_checking_nopointer_udata_code(index, type)
      end
    end

    def gen_checking_pointer_udata_code(index, type)
      expect_var[:index].to_be Integer
      expect_var[:type].to_be FFI::Clang::Type

      type_var = gen_name(:type)
      add_stmt("int #{type_var} = #{build_api('type', index)}")

      # フルユーザデータ型かを調べるコードを生成
      self << "if( #{type_var} == LUA_TUSERDATA ){" << endl
      self.next_level do
        # uservalueを取り出す
        call_api('getuservalue', index)

        # 型情報(文字列)を取り出す
        call_api('rawgeti', -1, userdata_type_index)

        typeinfo_var = gen_name('type')
        add_stmt("const char *#{typeinfo_var} = #{build_api('tostring', -1)}")

        self << "if( #{typeinfo_var} == NULL ){" << endl
        next_level do
          call_auxapi('error', '"missing type infomation"')
        end
        self << '}' << endl

        type = type.canonical_rec
        if type.pointer?
          if type.spelling == 'void *'
            # 型情報に'*'が含まれるかで判定
            self << "if( !strchr(#{typeinfo_var}, '*') ){" << endl
            next_level do
              call_auxapi('error', '"expected pointer type"')
            end
            self << '}' << endl
          else
            self << "if( strcmp(#{typeinfo_var}, \"#{type.spelling}\") && strcmp(#{typeinfo_var}, \"void *\") )"
            nest_block(true) do
              call_auxapi('error', "\"expected #{type.spelling}\"")
            end
          end
        else
          self << "if( strcmp(#{typeinfo_var}, \"#{type.spelling}\") ){" << endl
          next_level do
            call_auxapi('error', "\"invalid type\"")
          end
          self << '}' << endl
        end
        # uservalueと型情報をpop
        call_api('pop', 2)
      end
      self << '}' << endl

      self << "else if( #{type_var} != LUA_TNIL ) {" << endl
      next_level do
        call_auxapi('error', "\"expected userdata, but got %s\"", build_api('typename', build_api('type', index)))
      end
      self << '}' << endl
    end

    def gen_checking_nopointer_udata_code(index, type)
      expect_var[:index].to_be Integer
      expect_var[:type].to_be FFI::Clang::Type

      # フルユーザデータ型かを調べるコードを生成
      self << "if( #{build_api('type', index)} != LUA_TUSERDATA ){" << endl
      next_level do
        call_auxapi('error', "\"expected userdata, but got %s\"", build_api('typename', build_api('type', index)))
      end
      self << '}' << endl

      # uservalueを取り出す
      call_api('getuservalue', index)

      # 型情報(文字列)を取り出す
      call_api('rawgeti', -1, userdata_type_index)

      type_var = gen_name('type')
      add_stmt("const char *#{type_var} = #{build_api('tostring', -1)}")

      self << "if( #{type_var} == NULL ){" << endl
      next_level do
        call_auxapi('error', '"missing type infomation"')
      end
      self << '}' << endl

      type = type.canonical_rec
      if type.pointer?
        if type.spelling == 'void *'
          # 型情報に'*'が含まれるかで判定
          self << "if( !strchr(#{type_var}, '*') ){" << endl
          next_level do
            call_auxapi('error', '"expected pointer type"')
          end
          self << '}' << endl
        else
          self << "if( strcmp(#{type_var}, \"#{type.spelling}\") && strcmp(#{type_var}, \"void *\") )"
          nest_block(true) do
            call_auxapi('error', "\"expected #{type.spelling}\"")
          end
        end
      else
        self << "if( strcmp(#{type_var}, \"#{type.spelling}\") ){" << endl
        next_level do
          call_auxapi('error', "\"invalid type\"")
        end
        self << '}' << endl
      end

      # uservalueと型情報をpop
      call_api('pop', 2)
    end
  end
end
