require 'tempfile'
require 'optparse'
require 'stringio'

module Luca
  class Compiler
    include Varspec
    include Varspec::BuiltinMatcher

    # @return [String] 処理中のファイル名
    attr_reader :filename

    # @return [LocationMap]
    attr_reader :loc_map

    PREPROCESS = 0
    PARSE = 1
    DESUGAR = 2
    ANALYZE = 3
    GENERATE = 4

    def initialize(filename, phase: GENERATE, output: nil)
      expect_var[:filename].to_be String
      expect_var[:phase].to_be Integer
      expect_var[:output].to_be Or(String, StringIO, IO, nil)

      @filename = filename
      @phase = phase
      @output = output
    end

    def compile
      @error_list = []

      # プリプロセス
      preprocessed_filename = "#{filename}.i"

      # lua.hをインクルードしているかをチェック
      log = `clang -x c -MM -pedantic-errors #{filename}`
      log = log.split(' ')
      if %w(lua.h lualib.h lauxlib.h).any? { |h| log.include?(h) }
        fail 'explicitly lua header file included'
      end

      preprocessed_src, @loc_map = preprocess
      if @phase == PREPROCESS
        puts preprocessed_src
        return nil
      end

      File.open(preprocessed_filename, 'w') do |f|
        f.puts preprocessed_src
      end

      # 構文解析
      ast = Parser.parse(preprocessed_filename, filename, loc_map)
      flush_error

      FileUtils.rm(preprocessed_filename)

      if @phase == PARSE
        @output ||= "#{filename}.parsed.c"

        do_with_output_to(@output) do |f|
          Printer.new(filename, f, 2, loc_map).print_ast(ast)
        end
        return
      end

      # desugar
      ast = DesugarContext.new.desugar_program(ast)
      flush_error

      if @phase == DESUGAR
        @output ||= "#{filename}.desugared.c"
        do_with_output_to(@output) do |f|
          Printer.new(f, 2, loc_map).print_ast(ast)
          return
        end
      end

      # analyze
      ast = Analyzer.new(self).analyze(ast)
      flush_error

      # コード生成
      @output ||= "#{filename}.c"
      macro_file = "#{filename}.macro.h"
      do_with_output_to(@output) do |f|
        File.open(macro_file, 'w') do |tmp|
          tmp.puts '#include <lua.h>'
          tmp.puts '#include <lualib.h>'
          tmp.puts '#include <lauxlib.h>'
        end
        f.puts 'int strcmp(const char *, const char *);'
        f.puts '#define NULL ((void *)0)'
        f.puts `gcc -x c -E -dD #{macro_file} | grep -i "#define lua"`
        FileUtils.rm(macro_file)
        CodeGenerator.new(self, filename, f, 2, loc_map).gen_code(ast)
      end
    end

    def add_error(line, msg)
      loc = loc_map.logical_loc_of(line)
      msg = "#{loc.filename}:#{loc.line}: #{msg}"
      @error_list << msg
    end

    def flush_error(*err)
      fail ArgumentError, 'argument size must be even number' if err.size.odd?

      err.each_slice(2) do |pair|
        add_error(pair[0], pair[1])
      end

      return if @error_list.empty?

      @error_list.each do |e|
        $stderr.puts e
      end

      e = CompileError.create_by(@error_list)
      fail e
    end

    def preprocess
      preprocessed_src = `clang -x c -E -std=c99 -pedantic-errors -Wall #{filename}`

      src = ''
      loc_list = []
      preprocessed_src.each_line.with_index(1) do |str, line|
        if str =~ /^#\s*(\d+)\s+"(.+)"(.*)$/
          line_info = [line, Regexp.last_match[1].to_i,
                       Regexp.last_match[2], Regexp.last_match[3].chomp]
          loc_list << line_info
          src << "\n"
        else
          src << str
        end
      end

      [src, LocationMap.new(filename, loc_list)]
    end
    private :preprocess

    def do_with_output_to(output, *args)
      expect_var[:output].to_be Or(String, StringIO, IO)

      if output.is_a? String
        File.open(output, 'w') do |file|
          yield file, *args
        end
      else
        yield output, *args
      end
    end
    private :do_with_output_to

    def self.compile(filename, phase: GENERATE, output: nil)
      self.new(filename, phase: phase, output: output).compile
    end

    def self.run(args = [])
      opt = OptionParser.new
      output = nil
      opt.on('-o NAME') do |v|
        output = v
      end

      phase = GENERATE
      opt.on('-E') do |v|
        phase = PREPROCESS
      end

      opt.parse!(args)

      if args.length != 1
        e = CompileError.new('specify a source LuCa file')
        fail e
      end

      filename = args.first
      c_source = "#{filename}.c"
      output ||= 'a.out'
      if output == '-'
        unless filename.endwith?('.luca')
          e = CompileError.new('filename should end with `.luca\'')
          fail e
        end
        filename.match(/(.+)\.luca$/) do |md|
          output = md[1]
        end
      end
      self.compile(filename, phase: phase, output: c_source)
      return nil if phase == PREPROCESS
    end
  end
end
