require 'singleton'
require 'tempfile'

module Luca
  # ASTを印字するためのクラス
  #
  # @author autopp <autopp.inc@gmail.com>
  #
  class Printer
    include Varspec::BuiltinMatcher

    attr_reader :current_line

    def initialize(filename, output, width, loc_map)
      expect_var[:filename].of_instance_to_be String
      expect_var[:output].of_instance_to_be Or(IO, Tempfile)
      expect_var[:width].of_instance_to_be Integer
      expect_var[:loc_map].of_instance_to_be LocationMap
      @with_lmacro = true
      @current_line = 1
      @tab = ' ' * width
      @level = 0
      @is_newline = true
    end

    def with_lmacro?
      @with_lmacro != false
    end

    def logical_loc_of(line)
      @loc_map.logical_loc_of(line)
    end

    def current_logical_loc
      logical_loc_of(@current_line)
    end

    def <<(obj)
      case obj
      when Ast
        self.print_ast(obj)
      when Array
        obj.each do |o|
          self << o
        end
      when EndLine
        self.newline
      when List
        self.print_ast_list(obj.ls, obj.sep, obj.last_sep)
      when Token
        self.print_token(obj)
      else
        self.write_str(obj)
      end
      self
    end

    def write_str(obj, unindent = false)
      self.indent(unindent) if @is_newline
      @output << obj
      @is_newline = false
      self
    end

    def write_lmacro(actuial_line, next_line, filename = nil)
      expect_var[:actuial_line, :next_line].to_be Integer
      filename ||= @filename
      expect_var[:filename].to_be String

      filename = "\"#{filename}\"" if filename !~ /^".+"$/
      @current_line = actuial_line
      @output.puts "# #{next_line} #{filename}"
    end

    def newline
      if @with_lmacro
        newline_with_lmacro
      else
        newline_without_lmacro
      end
    end

    def newline_with_lmacro
      lmacro = @loc_map.find_lmacro(@current_line)
      if lmacro
        @output.puts lmacro
      else
        self.indent if @is_newline
        @output.puts ''
      end

      @current_line += 1
      @is_newline = true
      self
    end

    def newline_without_lmacro
      self.indent if @is_newline
      @output.puts ''

      @current_line += 1
      @is_newline = true
      self
    end

    def in_lmacro_mode
      old, @with_lmacro = @with_lmacro, true
      yield
      @with_lmacro = old
    end

    def in_nolmacro_mode
      old, @with_lmacro = @with_lmacro, false
      yield
      @with_lmacro = old
    end

    def indent(unindent = false)
      if unindent
        @output << (@tab * ([@level - 1, 0].max))
      else
        @output << (@tab * @level)
      end
    end

    def next_level(*args, sep: ', ', last_sep: false)
      @level += 1

      self.print_ast_list(args, sep, last_sep)
      yield if block_given?

      @level -= 1
      self
    end

    # ASTを出力する
    #
    # @param [Ast] ast 出力するAST
    #
    # @return [self]
    #
    def print_ast(ast)
      expect_var[:ast].to_be Ast

      ast.print_by(self)
      self
    end

    # リストを出力する
    #
    # @param [Array] list 出力するリスト
    # @param [Object] sep AST間の区切り文字列
    #
    # @return [self]
    #
    def print_ast_list(list, sep = ', ', last_sep = false)
      expect_var[:list].to_be Array

      unless list.empty?
        self << (list.first)
        list[1...list.length].each do |ast|
          self << sep
          self << ast
        end
        self << sep if last_sep
      end
      self
    end

    alias_method :print_list, :print_ast_list

    # トークンの出力
    #
    # @param [Token] token
    #
    # @return [self]
    #
    def print_token(token, unindent = false)
      token.newline_count.times do
        self.newline
      end
      self.write_str(token.value, unindent)
      self
    end

    # C++ライクな改行マニピュレータを実現するためのシングルトンクラス

    #
    # @author autopp <autopp.inc@gmail.com>
    #
    class EndLine
      include Singleton
    end

    ENDL = EndLine.instance
    def self.endl
      ENDL
    end

    # C++ライクなリストマニピュレータを実現するためのクラス
    #
    # @author autopp <autopp.inc@gmail.com>
    #
    class List
      include Varspec::BuiltinMatcher

      attr_reader :ls
      attr_reader :sep
      attr_reader :last_sep
      def initialize(ls, sep = ', ', last_sep = false)
        expect_var[:ls].of_instance_to_be Array
        @sep = sep
        expect_var[:last_sep].of_instance_to_be Boolean
      end
    end

    def self.list(ls, sep = ', ', last_sep = false)
      List.new(ls, sep, last_sep)
    end
  end
end
